# Copyright 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Wrappers to parse the output of the PCM JSON data from VIOS."""

import abc
import json

import six

from pypowervm.wrappers import pcm


class ViosInfo(object):
    """Represents a monitor sample from the Virtual I/O Server monitor.

    The VIOS PCM monitor JSON data can be parsed by this.  The base structure
    is:
    - ViosInfo
      - Info
      - ViosSample
        - ViosMemory
        - ViosNetwork
          - ViosNetworkAdpt (List)
          - ViosSharedEthernetAdapter (List)
        - ViosStorage
          - ViosFCPhysAdpt (List)
            - ViosFCVirtAdpt (List)
          - ViosStoragePAdpt (List)
          - ViosStorageVAdpt (List)
          - ViosSSP (List)
    """

    def __init__(self, raw_json):
        data = json.loads(raw_json)
        systemUtil = data.get('systemUtil')
        self.info = pcm.Info(systemUtil.get('utilInfo'))
        self.sample = ViosSample(systemUtil.get('utilSample'))


class ViosSample(object):

    def __init__(self, util_sample):
        self.time_stamp = util_sample.get('timeStamp')

        # TODO(thorst) Evaluate with multi VIOS.
        vios = util_sample.get('viosUtil')[0]
        # Convert the ID to keep consistent with phyp.
        self.id = int(vios.get('id'))
        self.name = vios.get('name')

        # Complex types
        mem = vios.get('memory')
        self.mem = ViosMemory(mem) if mem else None

        net = vios.get('network')
        self.network = ViosNetwork(net) if net else None

        storage = vios.get('storage')
        self.storage = ViosStorage(storage) if storage else None


class ViosMemory(object):

    def __init__(self, mem):
        self.utilized_mem = mem.get('utilizedMem')


class ViosNetwork(object):
    """The Network elements within the VIOS."""

    def __init__(self, net):
        self.adpts = [ViosNetworkAdpt(x) for x in net.get('genericAdapters')]
        self.seas = [ViosSharedEthernetAdapter(x)
                     for x in net.get('sharedAdapters')]


class ViosNetworkAdpt(object):
    """Represents a Network Adapter on the system."""

    def __init__(self, adpt):
        self.name = adpt.get('id')
        # Type: 'virtual' or 'physical' or 'sea' (if NetworkBridge)
        self.type = adpt.get('type')
        self.physical_location = adpt.get('physicalLocation')
        self.received_packets = adpt.get('receivedPackets')
        self.sent_packets = adpt.get('sentPackets')
        self.dropped_packets = adpt.get('droppedPackets')
        self.received_bytes = adpt.get('receivedBytes')
        self.sent_bytes = adpt.get('sentBytes')


class ViosSharedEthernetAdapter(ViosNetworkAdpt):
    """Represents a Shared Ethernet Adapter on the VIOS."""

    def __init__(self, bridge):
        super(ViosSharedEthernetAdapter, self).__init__(bridge)
        self.bridged_adpts = bridge.get('bridgedAdapters')


class ViosStorage(object):
    """Represents the storage elements on the VIOS."""

    def __init__(self, storage):
        fc_adpts = storage.get('fiberChannelAdapters')
        self.fc_adpts = [ViosFCPhysAdpt(x) for x in fc_adpts]

        phys_adpts = storage.get('genericPhysicalAdapters')
        self.phys_adpts = [ViosStoragePAdpt(x) for x in phys_adpts]

        virt_adpts = storage.get('genericVirtualAdapters')
        self.virt_adpts = [ViosStorageVAdpt(x) for x in virt_adpts]

        ssps = storage.get('sharedStoragePools')
        self.ssps = [ViosSSP(x) for x in ssps]


@six.add_metaclass(abc.ABCMeta)
class ViosStorageAdpt(object):
    """Base class for storage adapters."""

    def __init__(self, adpt):
        self.name = adpt.get('id')
        self.physical_location = adpt.get('physicalLocation')
        self.num_reads = adpt.get('numOfReads')
        self.num_writes = adpt.get('numOfWrites')
        self.read_bytes = adpt.get('readBytes')
        self.write_bytes = adpt.get('writeBytes')


class ViosFCPhysAdpt(ViosStorageAdpt):
    """Represents a physical fiber channel adapter on the VIOS."""

    def __init__(self, adpt):
        super(ViosFCPhysAdpt, self).__init__(adpt)
        self.wwpn = adpt.get('wwpn')

        # Appears to be Gb/s interface speed
        self.running_speed = adpt.get('runningSpeed')

        # TODO(thorst) Add FC Ports (need vfc mappings)
        vadpts = adpt.get('ports')
        self.ports = [ViosFCVirtAdpt(x) for x in vadpts]


class ViosFCVirtAdpt(ViosStorageAdpt):
    """Represents a Virtual FC Port (NPIV)."""

    def __init__(self, vadpt):
        super(ViosFCVirtAdpt, self).__init__(vadpt)
        self.wwpn = vadpt.get('wwpn')

        # Appears to be Gb/s interface speed
        self.running_speed = vadpt.get('runningSpeed')


class ViosStoragePAdpt(ViosStorageAdpt):
    """Represents a physical storage adapter (typically a SAS drive)."""

    def __init__(self, adpt):
        super(ViosStoragePAdpt, self).__init__(adpt)
        self.type = adpt.get('type')


class ViosStorageVAdpt(ViosStorageAdpt):
    """Represents a virtual storage adapter (vscsi)."""

    def __init__(self, adpt):
        super(ViosStorageVAdpt, self).__init__(adpt)
        self.type = adpt.get('type')


class ViosSSP(object):
    """Represents a Shared Storage Pool (entire element)."""

    def __init__(self, ssp):
        self.name = ssp.get('id')
        self.pool_disks = ssp.get('poolDisks')
        self.num_reads = ssp.get('numOfReads')
        self.num_writes = ssp.get('numOfWrites')
        self.total_space = ssp.get('totalSpace')
        self.used_space = ssp.get('usedSpace')
        self.read_bytes = ssp.get('readBytes')
        self.write_bytes = ssp.get('writeBytes')



# Copyright 2016 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Tests for pypowervm.tasks.sriov."""

import mock
import testtools

import pypowervm.exceptions as ex
import pypowervm.tasks.sriov as tsriov
import pypowervm.tests.test_fixtures as fx
import pypowervm.tests.test_utils.test_wrapper_abc as twrap
import pypowervm.wrappers.iocard as card
import pypowervm.wrappers.managed_system as ms


def fake_sriov(mode, state, sriov_adap_id, phys_ports):
    return mock.Mock(mode=mode, state=state, sriov_adap_id=sriov_adap_id,
                     phys_loc_code='sriov_loc%d' % sriov_adap_id,
                     phys_ports=phys_ports)


def fake_pport(sriov_adap_id, port_id, cfg_lps, alloc_cap):
    return mock.Mock(sriov_adap_id=sriov_adap_id, port_id=port_id,
                     loc_code='pport_loc%d' % port_id,
                     min_granularity=float(port_id) / 1000,
                     cfg_max_lps=20,
                     cfg_lps=cfg_lps,
                     allocated_capacity=alloc_cap,
                     link_status=True)


def good_sriov(sriov_adap_id, pports):
    return fake_sriov(card.SRIOVAdapterMode.SRIOV,
                      card.SRIOVAdapterState.RUNNING, sriov_adap_id, pports)

ded_sriov = fake_sriov(card.SRIOVAdapterMode.DEDICATED, None, 86, [])
down_sriov = fake_sriov(card.SRIOVAdapterMode.SRIOV,
                        card.SRIOVAdapterState.FAILED, 68, [])


def sys_wrapper(sriovs, vnic_capable=True, vnic_failover_capable=True):
    mock_sys = mock.Mock(asio_config=mock.Mock(sriov_adapters=sriovs))
    mock_sys.get_capabilities.return_value = {
        'vnic_capable': vnic_capable,
        'vnic_failover_capable': vnic_failover_capable}
    return mock_sys


class TestSriov(testtools.TestCase):

    def setUp(self):
        super(TestSriov, self).setUp()
        self.adpt = self.useFixture(fx.AdapterFx()).adpt
        self.fake_sriovs = [
            good_sriov(1, [fake_pport(1, pid, lps, cap) for pid, lps, cap in (
                (11, 0, 0.95), (12, 9, 0.0), (13, 5, 0.03), (14, 20, 0.987))]),
            ded_sriov,
            good_sriov(2, [fake_pport(2, 21, 19, 0.3)]),
            down_sriov,
            good_sriov(3, []),
            good_sriov(4, [fake_pport(4, pid, 1, cap) for pid, cap in (
                (41, 0.02), (42, 0.01))]),
            good_sriov(5, [fake_pport(5, pid, lps, cap) for pid, lps, cap in (
                (51, 17, 0.49), (52, 3, 0.0), (53, 50, 0.05), (54, 11, 0.0),
                (55, 6, 0.4), (56, 13, 0.1), (57, 0, 0.15), (58, 7, 1.0))])]
        # Mark link status down on 5/55.
        self.fake_sriovs[6].phys_ports[4].link_status = False

    def test_get_good_sriovs(self):
        """Test _get_good_sriovs helper."""
        sriovs = tsriov._get_good_sriovs(self.fake_sriovs)
        self.assertEqual(5, len(sriovs))
        self.assertEqual(['sriov_loc%d' % x for x in range(1, 6)],
                         [sriov.phys_loc_code for sriov in sriovs])

        # Error case: none found.
        self.assertRaises(ex.NoRunningSharedSriovAdapters,
                          tsriov._get_good_sriovs, [ded_sriov, down_sriov])

    def test_get_good_pport_list(self):
        """Test _get_good_pport_list helper."""
        def validate_pports(pports, ids):
            # List of phys locs
            self.assertSetEqual({'pport_loc%d' % x for x in ids},
                                {pport.loc_code for pport in pports})
            # We added the appropriate sriov_adap_id
            for pport in pports:
                # Set up such that port ID xy always sits on adapter with ID x
                self.assertEqual(pport.port_id // 10, pport.sriov_adap_id)

        # Base case: no hits
        self.assertEqual([], tsriov._get_good_pport_list(
            self.fake_sriovs, ['nowt', 'to', 'see', 'here'], None, 0, False))
        # Validate redundancy - same thing but with nonzero redundancy
        self.assertRaises(
            ex.InsufficientSRIOVCapacity, tsriov._get_good_pport_list,
            self.fake_sriovs, ['nothing', 'to', 'see', 'here'], None, 1, False)
        # Make sure we can get the ones we specify, that are actually there.
        pports = tsriov._get_good_pport_list(
            self.fake_sriovs, ['pport_loc%d' % x for x in {
                51, 13, 68, 123, 21, 57, 42}], None, 4, False)
        validate_pports(pports, {42, 13, 57, 21, 51})
        # Make sure we can filter by saturation (capacity/LPs).  14, 53, and 58
        # should filter themselves - they're already too full for their
        # min_granularity and/or configured LPs.
        pports = tsriov._get_good_pport_list(
            self.fake_sriovs, ['pport_loc%d' % x for x in {
                58, 52, 14, 11, 53}], None, 0, False)
        validate_pports(pports, (52, 11))
        # Now specify capacity higher than 11 can handle - it should drop off
        pports = tsriov._get_good_pport_list(
            self.fake_sriovs, ['pport_loc%d' % x for x in {
                58, 52, 14, 11, 53}], 0.06, 0, False)
        validate_pports(pports, {52})
        # Filter link-down ports.  14, 53, and 58 don't appear because they're
        # saturated by capacity and/or configured LPs.  55 doesn't appear
        # because it's link-down.
        pports = tsriov._get_good_pport_list(
            self.fake_sriovs, ['pport_loc%d' % x for x in range(60)], None, 0,
            True)
        validate_pports(pports, {12, 52, 54, 42, 13, 41, 56, 57, 21, 51, 11})

    @mock.patch('pypowervm.wrappers.managed_system.System.get')
    def test_check_sys_vnic_capabilities(self, mock_sys_get):
        sys_yes_yes = sys_wrapper(None)
        sys_yes_no = sys_wrapper(None, vnic_failover_capable=False)
        sys_no_yes = sys_wrapper(None, vnic_capable=False)
        sys_no_no = sys_wrapper(None, vnic_capable=False,
                                vnic_failover_capable=False)
        # With sys param None, does a get; vnic & failover checks pass
        mock_sys_get.return_value = [sys_yes_yes]
        self.assertEqual(sys_yes_yes,
                         tsriov._check_sys_vnic_capabilities('adap', None, 2))
        mock_sys_get.assert_called_once_with('adap')

        mock_sys_get.reset_mock()

        # No get; vnic & !failover ok
        self.assertEqual(
            sys_yes_no,
            tsriov._check_sys_vnic_capabilities('adap', sys_yes_no, 1))
        mock_sys_get.assert_not_called()

        # Same, but 0 is a valid min
        self.assertEqual(
            sys_yes_no,
            tsriov._check_sys_vnic_capabilities('adap', sys_yes_no, 0))

        # vnic & !failover !ok
        self.assertRaises(
            ex.VNICFailoverNotSupportedSys,
            tsriov._check_sys_vnic_capabilities, 'adap', sys_yes_no, 2)

        # !vnic !ok even if failover ok (which would really never happen)
        self.assertRaises(
            ex.SystemNotVNICCapable,
            tsriov._check_sys_vnic_capabilities, 'adap', sys_no_yes, 2)

        # !vnic !failover !ok
        self.assertRaises(
            ex.SystemNotVNICCapable,
            tsriov._check_sys_vnic_capabilities, 'adap', sys_no_no, 1)

    @mock.patch('pypowervm.tasks.partition.get_active_vioses')
    def test_check_and_filter_vioses(self, mock_vioget):
        vios_yes_yes = mock.Mock(vnic_capable=True, vnic_failover_capable=True)
        vios_yes_no = mock.Mock(vnic_capable=True, vnic_failover_capable=False)
        vios_no_yes = mock.Mock(vnic_capable=False, vnic_failover_capable=True)
        vios_no_no = mock.Mock(vnic_capable=False, vnic_failover_capable=False)

        # No redundancy, no pre-seeded list.
        violist = [vios_yes_yes, vios_yes_no, vios_no_yes, vios_no_no]
        mock_vioget.return_value = violist
        # Because at least one VIOS was failover-capable, the non-capable one
        # is excluded.
        self.assertEqual([vios_yes_yes],
                         tsriov._check_and_filter_vioses('adap', None, 1))
        mock_vioget.assert_called_once_with('adap', xag=[], vios_wraps=None,
                                            find_min=1)

        mock_vioget.reset_mock()

        # With redundancy, pre-seeded list
        self.assertEqual([vios_yes_yes], tsriov._check_and_filter_vioses(
            'adap', violist, 2))
        mock_vioget.assert_called_once_with('adap', xag=[],
                                            vios_wraps=violist, find_min=1)

        # None capable
        violist = [vios_no_yes, vios_no_no]
        mock_vioget.return_value = violist
        self.assertRaises(ex.NoVNICCapableVIOSes,
                          tsriov._check_and_filter_vioses, 'adap', None, 1)

        # None redundancy capable
        violist = [vios_yes_no, vios_no_yes, vios_no_no]
        mock_vioget.return_value = violist
        self.assertRaises(ex.VNICFailoverNotSupportedVIOS,
                          tsriov._check_and_filter_vioses, 'adap', None, 2)

    @mock.patch('pypowervm.tasks.sriov._check_and_filter_vioses')
    @mock.patch('random.shuffle')
    def test_set_vnic_back_devs(self, mock_shuffle, mock_vioget):
        """Test set_vnic_back_devs."""
        mock_sys = sys_wrapper(self.fake_sriovs)
        mock_vioget.return_value = [mock.Mock(uuid='vios_uuid1'),
                                    mock.Mock(uuid='vios_uuid2'),
                                    mock.Mock(uuid='vios_uuid3')]
        self.adpt.build_href.side_effect = lambda *a, **k: '%s' % a[1]
        vnic = card.VNIC.bld(self.adpt, pvid=5)
        self.assertEqual(0, len(vnic.back_devs))
        # Silly case: redundancy of zero
        tsriov.set_vnic_back_devs(vnic, [], sys_w=mock_sys, redundancy=0)
        self.assertEqual(0, len(vnic.back_devs))
        mock_vioget.assert_called_once_with(self.adpt, None, 0)

        cap = 0.019
        # Things to note about the following:
        # - VIOSes rotate.  1, 2, 3, repeat.  If we hadn't mocked shuffle, the
        #   base order would be random, but they would still rotate in whatever
        #   that shuffled order was.
        # - The least-used (emptiest) physical ports come first...
        # - ...except (e.g. 21) we force distribution across cards, so...
        # - ...cards alternate until exhausted; hence 5 repeated at the end.
        # - Capacity set across the board according to the parameter.
        all_back_devs = [('vios_uuid1', 1, 12, cap),
                         ('vios_uuid2', 5, 52, cap),
                         ('vios_uuid3', 4, 42, cap),
                         ('vios_uuid1', 2, 21, cap),
                         ('vios_uuid2', 5, 54, cap),
                         ('vios_uuid3', 4, 41, cap),
                         ('vios_uuid1', 1, 13, cap),
                         ('vios_uuid2', 5, 56, cap),
                         ('vios_uuid3', 1, 11, cap),
                         ('vios_uuid1', 5, 57, cap),
                         ('vios_uuid2', 5, 55, cap),
                         ('vios_uuid3', 5, 51, cap)]
        # 5/55 is link-down.  When it drops off, the last one moves to VIOS 2.
        live_back_devs = all_back_devs[:10] + [('vios_uuid2', 5, 51, cap)]

        # Use 'em all
        tsriov.set_vnic_back_devs(vnic, ['pport_loc%d' % x for x in range(60)],
                                  sys_w=mock_sys, capacity=cap, redundancy=12)
        self.assertEqual(all_back_devs,
                         [(bd.vios_href, bd.sriov_adap_id, bd.pport_id,
                           bd.capacity) for bd in vnic.back_devs])

        # Check port status - 55 drops off
        tsriov.set_vnic_back_devs(vnic, ['pport_loc%d' % x for x in range(60)],
                                  sys_w=mock_sys, capacity=cap, redundancy=11,
                                  check_port_status=True)
        self.assertEqual(live_back_devs,
                         [(bd.vios_href, bd.sriov_adap_id, bd.pport_id,
                           bd.capacity) for bd in vnic.back_devs])

        # Fail if we can't satisfy redundancy
        self.assertRaises(
            ex.InsufficientSRIOVCapacity, tsriov.set_vnic_back_devs, vnic,
            ['pport_loc%d' % x for x in range(60)],
            sys_w=mock_sys, capacity=cap, redundancy=13)

        # The passed-in wrapper isn't modified if the method raises.
        self.assertEqual(live_back_devs,
                         [(bd.vios_href, bd.sriov_adap_id, bd.pport_id,
                           bd.capacity) for bd in vnic.back_devs])

        # Make sure redundancy caps it.
        # By reusing vnic without resetting its back_devs, we're proving the
        # documented behavior that the method clears first.
        tsriov.set_vnic_back_devs(vnic, ['pport_loc%d' % x for x in range(60)],
                                  sys_w=mock_sys, capacity=cap, redundancy=5)
        self.assertEqual(all_back_devs[:5],
                         [(bd.vios_href, bd.sriov_adap_id, bd.pport_id,
                           bd.capacity) for bd in vnic.back_devs])

        self.assertEqual(5, mock_shuffle.call_count)

    @mock.patch('pypowervm.wrappers.managed_system.System.get')
    def test_find_pports_for_portlabel(self, mock_sys_get):
        physnet = 'default'
        sriov_adaps = [
            mock.Mock(phys_ports=[
                mock.Mock(loc_code='port1', label='default'),
                mock.Mock(loc_code='port3', label='data1')]),
            mock.Mock(phys_ports=[
                mock.Mock(loc_code='port4', label='data2'),
                mock.Mock(loc_code='port2', label='default')])]
        sys = mock.Mock(asio_config=mock.Mock(sriov_adapters=sriov_adaps))
        mock_sys_get.return_value = [sys]
        pports = tsriov.find_pports_for_portlabel(physnet, sriov_adaps)
        self.assertEqual({'port1', 'port2'},
                         {pport.loc_code for pport in pports})

    @mock.patch('pypowervm.wrappers.managed_system.System.get')
    def test_find_pports_for_portlabel_blank(self, mock_sys_get):
        physnet = 'default'
        sriov_adaps = [
            mock.Mock(phys_ports=[
                mock.Mock(loc_code='port1', label=''),
                mock.Mock(loc_code='port3', label='data1')]),
            mock.Mock(phys_ports=[
                mock.Mock(loc_code='port4', label='data2'),
                mock.Mock(loc_code='port2', label='')])]
        sys = mock.Mock(asio_config=mock.Mock(sriov_adapters=sriov_adaps))
        mock_sys_get.return_value = [sys]
        pports = tsriov.find_pports_for_portlabel(physnet, sriov_adaps)
        self.assertEqual({'port1', 'port2'},
                         {pport.loc_code for pport in pports})


class TestSafeUpdatePPort(testtools.TestCase):

    @mock.patch('pypowervm.tasks.partition.get_partitions')
    @mock.patch('pypowervm.wrappers.iocard.VNIC.get')
    def test_get_lpar_vnics(self, mock_vnics, mock_get_pars):
        lpars = ['lpar1', 'lpar2', 'lpar3']
        mock_get_pars.return_value = lpars
        mock_vnics.side_effect = ['list1', 'list2', 'list3']
        self.assertEqual({'lpar%d' % i: 'list%d' % i for i in (1, 2, 3)},
                         tsriov.get_lpar_vnics('adap'))
        mock_get_pars.assert_called_once_with('adap', lpars=True, vioses=False)
        for lpar in lpars:
            mock_vnics.assert_any_call('adap', parent=lpar)

    def test_vnics_using_pport(self):
        lpar1 = mock.Mock()
        lpar1.configure_mock(name='lpar1', uuid='lpar_uuid1')
        lpar2 = mock.Mock()
        lpar2.configure_mock(name='lpar2', uuid='lpar_uuid2')
        vnic1 = mock.Mock(uuid='vnic_uuid1', back_devs=[
            mock.Mock(sriov_adap_id=1, pport_id=1),
            mock.Mock(sriov_adap_id=2, pport_id=2)])
        vnic2 = mock.Mock(uuid='vnic_uuid2', back_devs=[
            mock.Mock(sriov_adap_id=1, pport_id=2),
            mock.Mock(sriov_adap_id=2, pport_id=1)])
        vnic3 = mock.Mock(uuid='vnic_uuid3', back_devs=[
            mock.Mock(sriov_adap_id=3, pport_id=1),
            mock.Mock(sriov_adap_id=4, pport_id=2)])
        vnic4 = mock.Mock(uuid='vnic_uuid4', back_devs=[
            mock.Mock(sriov_adap_id=1, pport_id=2),
            mock.Mock(sriov_adap_id=4, pport_id=2)])
        lpar2vnics = {lpar1: [vnic1, vnic2], lpar2: [vnic3, vnic4]}
        # Not in use
        self.assertEqual([], tsriov._vnics_using_pport(mock.Mock(
            sriov_adap_id=1, port_id=3, loc_code='not_used'), lpar2vnics))
        # Used once
        ret = tsriov._vnics_using_pport(mock.Mock(
            sriov_adap_id=1, port_id=1, loc_code='loc1'), lpar2vnics)
        self.assertEqual(1, len(ret))
        # loc1 backs vNIC for LPAR lpar1 (lpar_uuid1 / vnic_uuid1)
        self.assertIn('loc1', ret[0])
        self.assertIn('lpar1', ret[0])
        self.assertIn('lpar_uuid1', ret[0])
        self.assertIn('vnic_uuid1', ret[0])
        # Used twice
        ret = tsriov._vnics_using_pport(mock.Mock(
            sriov_adap_id=1, port_id=2, loc_code='loc2'), lpar2vnics)
        self.assertEqual(2, len(ret))
        # Order of the return is not deterministic.  Reverse if necessary
        if 'lpar1' not in ret[0]:
            ret = ret[::-1]
        # loc2 backs vNIC for LPAR lpar1 (lpar_uuid1 / vnic_uuid2)
        self.assertIn('loc2', ret[0])
        self.assertIn('lpar1', ret[0])
        self.assertIn('lpar_uuid1', ret[0])
        self.assertIn('vnic_uuid2', ret[0])
        # loc2 backs vNIC for LPAR lpar2 (lpar_uuid2 / vnic_uuid4)
        self.assertIn('loc2', ret[1])
        self.assertIn('lpar2', ret[1])
        self.assertIn('lpar_uuid2', ret[1])
        self.assertIn('vnic_uuid4', ret[1])

    @mock.patch('pypowervm.tasks.sriov.get_lpar_vnics')
    @mock.patch('pypowervm.tasks.sriov._vnics_using_pport')
    def test_vet_port_usage(self, mock_vup, mock_glv):
        label_index = {'loc1': 'pre_label1', 'loc2': '', 'loc3': 'pre_label3',
                       'loc4': 'pre_label4'}
        # No LPs
        pport1 = mock.Mock(cfg_lps=0, loc_code='loc1', label='post_label1')
        # Pre-label empty, but label changed
        pport2 = mock.Mock(loc_code='loc2', label='post_label2')
        # Pre-label matches post-label
        pport3 = mock.Mock(loc_code='loc3', label='pre_label3')
        # Label changed
        pport4 = mock.Mock(loc_code='loc4', label='post_label4')
        pport5 = mock.Mock(loc_code='loc3', label='post_label3')
        # PPorts that hit the first three criteria (no LPs, label originally
        # unset, label unchanged) don't trigger expensive get_lpar_vnics or
        # _vnics_using_pport.
        sriov1 = mock.Mock(phys_ports=[pport1, pport2])
        sriov2 = mock.Mock(phys_ports=[pport3])
        ret = tsriov._vet_port_usage(
            mock.Mock(asio_config=mock.Mock(sriov_adapters=[sriov1, sriov2])),
            label_index)
        self.assertEqual([], ret)
        mock_vup.assert_not_called()
        mock_glv.assert_not_called()
        # Multiple pports that pass the easy criteria; get_lpar_vnics only
        # called once.
        mock_vup.side_effect = [1], [2]
        sriov3 = mock.Mock(phys_ports=[pport4, pport5])
        ret = tsriov._vet_port_usage(mock.Mock(
            adapter='adap', asio_config=mock.Mock(
                sriov_adapters=[sriov1, sriov3])), label_index)
        mock_glv.assert_called_once_with('adap')
        self.assertEqual([1, 2], ret)
        mock_vup.assert_has_calls([mock.call(pport4, mock_glv.return_value),
                                   mock.call(pport5, mock_glv.return_value)])

    @mock.patch('pypowervm.tasks.sriov._vet_port_usage')
    @mock.patch('pypowervm.tasks.sriov.LOG.warning')
    @mock.patch('fasteners.lock.ReaderWriterLock.write_lock')
    @mock.patch('pypowervm.wrappers.managed_system.System.getter')
    def test_safe_update_pports(self, mock_getter, mock_lock, mock_warn,
                                mock_vpu):
        mock_sys = mock.Mock(asio_config=mock.Mock(sriov_adapters=[
            mock.Mock(phys_ports=[mock.Mock(loc_code='loc1', label='label1'),
                                  mock.Mock(loc_code='loc2', label='label2')]),
            mock.Mock(phys_ports=[
                mock.Mock(loc_code='loc3', label='label3')])]))

        def changes_func(ret_bool):
            def changes(sys_w):
                mock_lock.assert_called()
                mock_lock.reset_mock()
                return ret_bool
            return changes

        # No force, no warnings, update requested
        mock_vpu.return_value = []
        self.assertEqual(mock_sys.update.return_value,
                         tsriov.safe_update_pports(
                             mock_sys, changes_func(True)))
        mock_warn.assert_not_called()
        # No force, no in-use, no update, use a getter: runs but doesn't update
        mock_sys.update.reset_mock()
        self.assertEqual(mock_getter.return_value, tsriov.safe_update_pports(
            ms.System.getter('adap'), changes_func(False)))
        mock_warn.assert_not_called()
        mock_getter.return_value.update.assert_not_called()
        # Update requested, some in-use, no force - raises
        mock_vpu.reset_mock()
        mock_vpu.return_value = [1]
        self.assertRaises(ex.CantUpdatePPortsInUse, tsriov.safe_update_pports,
                          mock_sys, changes_func(True), force=False)
        mock_warn.assert_not_called()
        mock_sys.update.assert_not_called()
        # Update requested, some in-use, force - runs & warns
        mock_vpu.reset_mock()
        mock_vpu.return_value = ['one', 'two']
        self.assertEqual(mock_sys.update.return_value,
                         tsriov.safe_update_pports(
                             mock_sys, changes_func(True), force=True))
        mock_warn.assert_has_calls([mock.call(mock.ANY), mock.call('one'),
                                    mock.call('two')])


class TestMisc(twrap.TestWrapper):
    file = 'sys_with_sriov.txt'
    wrapper_class_to_test = ms.System

    def test_find_pport(self):
        self.assertIsNone(tsriov.find_pport(self.dwrap, 'bogus'))
        pport = tsriov.find_pport(self.dwrap, 'U78C7.001.RCH0004-P1-C8-T2')
        self.assertEqual('U78C7.001.RCH0004-P1-C8',
                         pport.sriov_adap.phys_loc_code)
        self.assertEqual(1, pport.sriov_adap_id)
        # It's a converged port...
        self.assertIsInstance(pport, card.SRIOVConvPPort)
        # ...which is also an ethernet port
        self.assertIsInstance(pport, card.SRIOVEthPPort)
        self.assertEqual('U78C7.001.RCH0004-P1-C8-T2', pport.loc_code)
        self.assertEqual(1, pport.port_id)



# Copyright 2014, 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import unittest

import six

import pypowervm.exceptions as pvmex

msg_params = {
    "backing_dev": "backing_dev_param",
    "cpu_size": 678,
    "element": "element_param",
    "element_type": "element_type_param",
    "error": "error_param",
    "file_name": "file_name_param",
    "image_volume": "image_volume_param",
    "lpar_name": "lpar_name_param",
    "min_vios": "min_vios_param",
    "name": "name_param",
    "operation_name": "operation_name_param",
    "reason": "reason_param",
    "seconds": 147,
    "valid_values": "valid_values_param",
    "vios_state": "vios_state_param",
    "volume": "volume_param",
    "access_file": "testfile",
}

os.environ['LANG'] = 'en_US'

class2msg = {
    pvmex.NotFound:
    "Element not found: element_type_param element_param",
    pvmex.LPARNotFound:
    "LPAR not found: lpar_name_param",
    pvmex.JobRequestFailed:
    "The 'operation_name_param' operation failed. error_param",
    pvmex.JobRequestTimedOut:
    "The 'operation_name_param' operation failed. "
    "Failed to complete the task in 147 seconds.",
    pvmex.AuthFileReadError:
    "OS denied access to file testfile.",
    pvmex.AuthFileAccessError:
    "OS encountered an I/O error attempting to read file testfile: "
    "error_param",
    pvmex.MigrationFailed:
    "The migration task failed. error_param"
}


class TestExceptions(unittest.TestCase):
    """Test coverage for the pypowervm.exceptions module."""

    def raise_helper(self, e):
        raise e

    def fmt_helper(self, eclass, expected_message):
        e = eclass(**msg_params)
        self.assertRaises(eclass, self.raise_helper, e)
        try:
            raise e
        except eclass as e1:
            self.assertEqual(e1.args[0], expected_message)

    def test_Error(self):
        e = pvmex.Error("test")
        self.assertRaises(pvmex.Error, self.raise_helper, e)
        try:
            raise e
        except pvmex.Error as e1:
            self.assertEqual(e1.args[0], "test")

    def test_fmterrors(self):
        for e, s in six.iteritems(class2msg):
            try:
                self.fmt_helper(e, s)
            except ValueError:
                self.fail(s)

    def test_bogusformatparams(self):
        class Bogus(pvmex.AbstractMsgFmtError):
            msg_fmt = "This has a %(bogus)s format parameter."

        self.assertRaises(KeyError, Bogus, **msg_params)

if __name__ == "__main__":
    unittest.main()



# Copyright 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
import six
import testtools

import pypowervm.entities as ent
import pypowervm.exceptions as pexc
from pypowervm.tasks import vterm
import pypowervm.tests.test_fixtures as fx


class TestVterm(testtools.TestCase):
    """Unit Tests for LPAR vterm."""

    def setUp(self):
        super(TestVterm, self).setUp()
        self.adpt = self.useFixture(
            fx.AdapterFx(traits=fx.LocalPVMTraits)).adpt

    @mock.patch('pypowervm.wrappers.job.Job.run_job')
    def test_close_vterm_non_local(self, mock_run_job):
        """Performs a close LPAR vterm test."""
        mock_resp = mock.MagicMock()
        mock_resp.entry = ent.Entry(
            {}, ent.Element('Dummy', self.adpt), self.adpt)
        self.adpt.read.return_value = mock_resp
        vterm._close_vterm_non_local(self.adpt, '12345')
        self.assertEqual(1, mock_run_job.call_count)
        self.assertEqual(1, self.adpt.read.call_count)
        # test exception path
        mock_run_job.side_effect = pexc.LPARNotFound(
            lpar_name='12345')
        self.assertRaises(pexc.LPARNotFound,
                          vterm._close_vterm_non_local, self.adpt, '12345')
        mock_run_job.reset_mock()

    @mock.patch('pypowervm.tasks.vterm._get_lpar_id')
    @mock.patch('pypowervm.tasks.vterm._run_proc')
    def test_open_vnc_vterm(self, mock_run_proc, mock_get_lpar_id):
        """Validates the output from the mkvterm if a vterm is not active."""
        mock_get_lpar_id.return_value = '4'
        std_out = '5903'
        std_err = ('VNC is started on port 5903 for localhost access '
                   'only.  Use \'rmvterm --id 4\' to close it.')
        mock_run_proc.return_value = (0, std_out, std_err)

        resp = vterm.open_localhost_vnc_vterm(self.adpt, 'lpar_uuid')

        mock_run_proc.assert_called_once_with(['mkvterm', '--id', '4', '--vnc',
                                               '--local'])
        self.assertEqual(5903, resp)

    @mock.patch('pypowervm.tasks.vterm._get_lpar_id')
    @mock.patch('pypowervm.tasks.vterm._run_proc')
    def test_open_vnc_vterm_existing(self, mock_run_proc, mock_get_lpar_id):
        """Validates the output from the mkvterm if a VNC vterm is active."""
        mock_get_lpar_id.return_value = '4'
        std_out = '5903'
        std_err = ('\nVNC server is already started on port 5903. Use '
                   '\'rmvterm --id 4\' to close it.')
        mock_run_proc.return_value = (3, std_out, std_err)

        resp = vterm.open_localhost_vnc_vterm(self.adpt, 'lpar_uuid')

        mock_run_proc.assert_called_once_with(['mkvterm', '--id', '4', '--vnc',
                                               '--local'])
        self.assertEqual(5903, resp)

    @mock.patch('pypowervm.tasks.vterm.close_vterm')
    @mock.patch('pypowervm.tasks.vterm._get_lpar_id')
    @mock.patch('pypowervm.tasks.vterm._run_proc')
    def test_open_vnc_vterm_nonvnc_force(self, mock_run_proc,
                                         mock_get_lpar_id, mock_close):
        """Validates the output from mkvterm if non-vnc active and force."""
        mock_get_lpar_id.return_value = '4'
        std_out_1 = ""
        std_err_1 = ("The vterm is currently in use by process 120352.  "
                     "Use 'rmvterm --id 4' to close it.")
        std_out_2 = '5903'
        std_err_2 = ('VNC is started on port 5903 for localhost access '
                     'only.  Use \'rmvterm --id 4\' to close it.')
        mock_run_proc.side_effect = [(3, std_out_1, std_err_1),
                                     (0, std_out_2, std_err_2)]

        resp = vterm.open_localhost_vnc_vterm(self.adpt, 'lpar_uuid',
                                              force=True)

        # Validation
        mock_close.assert_called_once_with(self.adpt, 'lpar_uuid')
        mock_run_proc.assert_any_call(['mkvterm', '--id', '4', '--vnc',
                                       '--local'])
        self.assertEqual(2, mock_run_proc.call_count)
        self.assertEqual(5903, resp)

    @mock.patch('pypowervm.tasks.vterm.close_vterm')
    @mock.patch('pypowervm.tasks.vterm._get_lpar_id')
    @mock.patch('pypowervm.tasks.vterm._run_proc')
    def test_open_vnc_vterm_nonvnc_noforce(self, mock_run_proc,
                                           mock_get_lpar_id, mock_close):
        """Validates the output from mkvterm if non-vnc active and no force."""
        mock_get_lpar_id.return_value = '4'
        std_out = ""
        std_err = ("The vterm is currently in use by process 120352.  "
                   "Use 'rmvterm --id 4' to close it.")
        mock_run_proc.return_value = (3, std_out, std_err)

        self.assertRaises(pexc.VNCBasedTerminalFailedToOpen,
                          vterm.open_localhost_vnc_vterm, self.adpt,
                          'lpar_uuid')

        # Validation
        mock_close.assert_not_called()
        mock_run_proc.assert_called_with(['mkvterm', '--id', '4', '--vnc',
                                          '--local'])
        self.assertEqual(1, mock_run_proc.call_count)

    @mock.patch('pypowervm.tasks.vterm.close_vterm')
    @mock.patch('pypowervm.tasks.vterm._get_lpar_id')
    @mock.patch('pypowervm.tasks.vterm._run_proc')
    def test_open_vnc_vterm_force_bad_error(self, mock_run_proc,
                                            mock_get_lpar_id, mock_close):
        """Validates the output from mkvterm if force but unexpected error."""
        mock_get_lpar_id.return_value = '4'
        std_out = ""
        std_err = ("The mkvterm command failed for an unexpected reason")
        mock_run_proc.return_value = (2, std_out, std_err)

        self.assertRaises(pexc.VNCBasedTerminalFailedToOpen,
                          vterm.open_localhost_vnc_vterm, self.adpt,
                          'lpar_uuid', force=True)

        # Validation
        mock_close.assert_not_called()
        mock_run_proc.assert_called_with(['mkvterm', '--id', '4', '--vnc',
                                          '--local'])
        self.assertEqual(1, mock_run_proc.call_count)

    @mock.patch('pypowervm.tasks.vterm._get_lpar_id')
    @mock.patch('pypowervm.tasks.vterm._run_proc')
    def test_close_vterm_local(self, mock_run_proc, mock_get_lpar_id):
        mock_get_lpar_id.return_value = '2'
        vterm._close_vterm_local(self.adpt, '5')
        mock_run_proc.assert_called_once_with(['rmvterm', '--id', '2'])


class TestVNCSocketListener(testtools.TestCase):
    """Unit Tests for _VNCSocketListener vterm."""

    def setUp(self):
        super(TestVNCSocketListener, self).setUp()
        self.adpt = self.useFixture(
            fx.AdapterFx(traits=fx.LocalPVMTraits)).adpt

        self.srv = vterm._VNCSocketListener(
            self.adpt, '5901', '1.2.3.4', True, remote_ips=['1.2.3.5'])
        self.rptr = vterm._VNCRepeaterServer(self.adpt, 'uuid', '5800')

        vterm._VNC_LOCAL_PORT_TO_REPEATER['5800'] = self.rptr
        vterm._VNC_PATH_TO_UUID['path'] = 'uuid'
        vterm._VNC_PATH_TO_UUID['test'] = 'uuid'
        vterm._VNC_UUID_TO_LOCAL_PORT['uuid'] = '5800'

    def tearDown(self):
        """Tear down the Session instance."""

        vterm._VNC_PATH_TO_UUID['path'] = None
        vterm._VNC_PATH_TO_UUID['test'] = None
        vterm._VNC_UUID_TO_LOCAL_PORT['1.2.3.4'] = None
        vterm._VNC_LOCAL_PORT_TO_REPEATER['5800'] = None
        self.rptr = None

        super(TestVNCSocketListener, self).tearDown()

    def test_stop(self):
        self.assertTrue(self.srv.alive)
        self.srv.stop()
        self.assertFalse(self.srv.alive)

    @mock.patch('select.select')
    @mock.patch('socket.socket')
    def test_new_client(self, mock_sock, mock_select):
        mock_srv = mock.MagicMock()
        mock_s_sock, mock_c_sock = mock.MagicMock(), mock.MagicMock()
        mock_sock.return_value = mock_s_sock
        mock_select.return_value = [mock_c_sock], None, None
        mock_srv.accept.return_value = mock_c_sock, ('1.2.3.5', '40675')
        mock_c_sock.recv.return_value = "CONNECT path HTTP/1.8\r\n\r\n"

        self.srv._new_client(mock_srv)

        mock_c_sock.sendall.assert_called_once_with(
            "HTTP/1.8 200 OK\r\n\r\n")
        mock_s_sock.connect.assert_called_once_with(('127.0.0.1', '5800'))

        self.assertEqual({mock_c_sock: mock_s_sock, mock_s_sock: mock_c_sock},
                         self.rptr.peers)

    def test_check_http_connect(self):
        sock = mock.MagicMock()
        sock.recv.return_value = "INVALID"
        uuid, http_code = self.srv._check_http_connect(sock)
        self.assertIsNone(uuid)
        self.assertEqual('1.1', http_code)

        # Test a good string
        sock.reset_mock()
        sock.recv.return_value = 'CONNECT test HTTP/2.0\r\n\r\n'
        uuid, http_code = self.srv._check_http_connect(sock)
        self.assertEqual('uuid', uuid)
        self.assertEqual('2.0', http_code)

    def test_new_client_bad_ip(self):
        """Tests that a new client will be rejected if a bad IP."""
        mock_srv = mock.MagicMock()
        mock_c_sock = mock.MagicMock()
        mock_srv.accept.return_value = mock_c_sock, ('1.2.3.8', '40675')

        self.srv._new_client(mock_srv)

        self.assertEqual(self.rptr.peers, {})
        self.assertEqual(1, mock_c_sock.close.call_count)

    @mock.patch('select.select')
    def test_new_client_validation_checks(self, mock_select):
        mock_srv = mock.MagicMock()
        mock_c_sock = mock.MagicMock()
        mock_select.return_value = None, None, None
        mock_srv.accept.return_value = mock_c_sock, ('1.2.3.5', '40675')

        # This mock has no 'socket ready'.
        self.srv._new_client(mock_srv)
        self.assertEqual(self.rptr.peers, {})
        mock_c_sock.sendall.assert_called_with(
            "HTTP/1.1 400 Bad Request\r\n\r\n")
        self.assertEqual(1, mock_c_sock.close.call_count)

        # Reset the select so that the validation check fails
        mock_c_sock.reset_mock()
        mock_select.return_value = [mock_c_sock], None, None
        mock_c_sock.recv.return_value = 'bad_check'
        self.srv._new_client(mock_srv)
        self.assertEqual(self.rptr.peers, {})
        mock_c_sock.sendall.assert_called_with(
            "HTTP/1.1 400 Bad Request\r\n\r\n")
        self.assertEqual(1, mock_c_sock.close.call_count)

    @mock.patch('pypowervm.tasks.vterm._close_vterm_local')
    def test_close_client(self, mock_close):
        client, server = mock.Mock(), mock.Mock()
        self.rptr.add_socket_connection_pair(client, server)
        with mock.patch('time.sleep'):
            self.rptr._close_client(client)

            self.assertTrue(client.close.called)
            self.assertTrue(server.close.called)
            self.assertEqual({}, self.rptr.peers)

            # Get the killer off the thread and wait for it to complete.
            self.rptr.vnc_killer.join()
            mock_close.assert_called_once_with(self.adpt, 'uuid')

    @mock.patch('pypowervm.tasks.vterm._VNCSocketListener._new_client')
    @mock.patch('select.select')
    @mock.patch('socket.socket')
    def test_run_new_client(self, mock_socket, mock_select, mock_new_client):
        mock_server = mock.MagicMock()
        mock_socket.return_value = mock_server
        mock_server.count = 0

        # Used to make sure we don't loop indefinitely.
        bad_select = mock.MagicMock()

        # Since we are only listening on the server socket, we will have it
        # return that same socket the first 2 times, and then also return a
        # bad socket just to make sure we appropriately broke out before then
        mock_select.side_effect = [([mock_server], [], []),
                                   ([mock_server], [], []),
                                   ([bad_select], [], [])]

        def new_client(server):
            # Keep track of a counter so we make sure we break out of the
            # loop after the 2nd call to make sure it works twice at least
            if server.count == 1:
                self.srv.alive = False
            # We are setting alive to false after the 2nd
            # so we want to make sure it doesn't get here
            if server == bad_select:
                raise Exception('Invalid Loop Call')
            server.count += 1

        mock_new_client.side_effect = new_client

        # If this runs...we've pretty much validated.  Because it will end.
        # If it doesn't end...the fail in the 'select' side effect shoudl catch
        # it.
        self.srv.run()

        # Make sure the close was called on all of the sockets.
        self.assertTrue(mock_server.close.called)

        # Make sure the select was called with a timeout.
        mock_select.assert_called_with(mock.ANY, mock.ANY, mock.ANY, 1)

    @mock.patch('select.select')
    @mock.patch('ssl.wrap_socket', mock.Mock())
    def test_enable_x509_authentication(self, mock_select):
        mock_select.return_value = [mock.Mock()], None, None
        csock, ssock = _FakeSocket(), _FakeSocket()
        ssock.recv_buffer = b'RFB 003.008\n\x01\x01'
        csock.recv_buffer = b'RFB 003.007\n\x13\x00\x02\x00\x00\x01\x04'
        # Test the method to do the handshake to enable VeNCrypt Authentication
        self.srv.set_x509_certificates('cacert1', 'cert1', 'key1')
        nsock = self.srv._enable_x509_authentication(csock, ssock)
        # Make sure that we didn't get an error and the TLS Socket was created
        self.assertIsNotNone(nsock, 'The TLS Socket was not created')
        # Verify that the data sent to the Client Socket matches expected
        csocksnd = b'RFB 003.008\n\x01\x13\x00\x02\x00\x01\x00\x00\x01\x04\x01'
        self.assertEqual(csock.send_buffer, csocksnd,
                         'Did not send to the client socket what was expected')
        # Verify that the data sent to the Server Socket matches expected
        self.assertEqual(ssock.send_buffer, b'RFB 003.007\n\x01',
                         'Did not send to the server socket what was expected')

    @mock.patch('select.select')
    def test_enable_x509_authentication_bad_auth_type(self, mock_select):
        mock_select.return_value = [mock.Mock()], None, None
        csock, ssock = _FakeSocket(), _FakeSocket()
        ssock.recv_buffer = b'RFB 003.008\n\x01\x01'
        csock.recv_buffer = b'RFB 003.007\n\x14\x00\x02\x00\x00\x01\x04'
        # Test the method to do the handshake to enable VeNCrypt Authentication
        self.srv.set_x509_certificates('cacert1', 'cert1', 'key1')
        nsock = self.srv._enable_x509_authentication(csock, ssock)
        # Make sure that we got an error and it didn't create the TLS Socket
        self.assertIsNone(nsock, 'Expected an error validating auth type')
        # Verify that the data sent to the Client Socket matches expected
        csocksnd = b'RFB 003.008\n\x01\x13\x00\x02\x01'
        self.assertEqual(csock.send_buffer, csocksnd,
                         'Did not send to the client socket what was expected')

    @mock.patch('select.select')
    def test_enable_x509_authentication_bad_auth_version(self, mock_select):
        mock_select.return_value = [mock.Mock()], None, None
        csock, ssock = _FakeSocket(), _FakeSocket()
        ssock.recv_buffer = b'RFB 003.008\n\x01\x01'
        csock.recv_buffer = b'RFB 003.007\n\x13\x00\x01\x00\x00\x01\x04'
        # Test the method to do the handshake to enable VeNCrypt Authentication
        self.srv.set_x509_certificates('cacert1', 'cert1', 'key1')
        nsock = self.srv._enable_x509_authentication(csock, ssock)
        # Make sure that we got an error and it didn't create the TLS Socket
        self.assertIsNone(nsock, 'Expected an error validating auth version')
        # Verify that the data sent to the Client Socket matches expected
        csocksnd = b'RFB 003.008\n\x01\x13\x00\x02\x01'
        self.assertEqual(csock.send_buffer, csocksnd,
                         'Did not send to the client socket what was expected')

    @mock.patch('select.select')
    def test_enable_x509_authentication_bad_auth_subtype(self, mock_select):
        mock_select.return_value = [mock.Mock()], None, None
        csock, ssock = _FakeSocket(), _FakeSocket()
        ssock.recv_buffer = b'RFB 003.008\n\x01\x01'
        csock.recv_buffer = b'RFB 003.007\n\x13\x00\x02\x00\x00\x01\x03'
        # Test the method to do the handshake to enable VeNCrypt Authentication
        self.srv.set_x509_certificates('cacert1', 'cert1', 'key1')
        nsock = self.srv._enable_x509_authentication(csock, ssock)
        # Make sure that we got an error and it didn't create the TLS Socket
        self.assertIsNone(nsock, 'Expected an error validating auth sub-type')
        # Verify that the data sent to the Client Socket matches expected
        csocksnd = b'RFB 003.008\n\x01\x13\x00\x02\x00\x01\x00\x00\x01\x04\x00'
        self.assertEqual(csock.send_buffer, csocksnd,
                         'Did not send to the client socket what was expected')


class _FakeSocket(object):

    def __init__(self):
        self.recv_buffer, self.send_buffer = b'', b''
        self.recv_bytes, self.send_bytes = 0, 0

    def recv(self, bufsize):
        bufsize = bufsize if isinstance(bufsize, int) else ord(bufsize)
        chunk = self.recv_buffer[self.recv_bytes:self.recv_bytes+bufsize]
        if not isinstance(chunk, six.binary_type):
            chunk = six.binary_type(chunk, 'utf-8')
        self.recv_bytes += bufsize
        return chunk

    def sendall(self, string):
        if not isinstance(string, six.binary_type):
            string = six.binary_type(string, 'utf-8')
        self.send_buffer += string
        self.send_bytes += len(string)



# Copyright 2014, 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""This Adapter helper retries a request when 'VIOS busy' error is detected.

A 'VIOS busy' error usually means the VIOS is processing another operation
which is mutually exclusive with the submitted request.  If this state
persists, it may mean the VIOS is in need of manual intervention.
"""

import time

import pypowervm.const as c
import pypowervm.exceptions as pvmex
import pypowervm.wrappers.entry_wrapper as ew
import pypowervm.wrappers.http_error as he

# Make UT a little easier
SLEEP = time.sleep


def vios_busy_retry_helper(func, max_retries=3, delay=5):
    """This helper retries the request if the resource is busy.

    Args:
        func: The Adapter request method to call
        max_retries (int): Max number retries.
    """

    def is_retry(http_error):
        """Determines if the error is one that can be retried."""
        # If for some reason it is not an Http Error, it can't be retried.
        if not isinstance(http_error, he.HttpError):
            return False

        # Check if the VIOS is clearly busy, or if there is a service
        # unavailable.  The service unavailable can occur on child objects when
        # a VIOS is busy.
        return (http_error.is_vios_busy() or
                http_error.status == c.HTTPStatus.SERVICE_UNAVAILABLE)

    def wrapper(*args, **kwds):
        retries = 0
        while True:
            try:
                # Call the request()
                resp = func(*args, **kwds)
            except pvmex.Error as e:
                # See if there are any retries left.
                if retries >= max_retries:
                    raise

                # See if the system was busy
                resp = e.response
                if resp and resp.body and resp.entry:
                    wrap = ew.EntryWrapper.wrap(resp.entry)
                    if is_retry(wrap):
                        retries += 1
                        # Wait a few seconds before trying again, scaling
                        # out the delay based on the retry count.
                        SLEEP(delay * retries)
                        continue

                # Doesn't look like a VIOS busy error, so just raise it.
                raise
            else:
                return resp

    return wrapper



# Copyright 2014, 2017 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import unittest

from pypowervm.tests.test_utils import pvmhttp
import pypowervm.wrappers.managed_system as ms
import pypowervm.wrappers.mtms as mtmwrap

_MS_HTTPRESP_FILE = "managedsystem.txt"
_MC_HTTPRESP_FILE = "managementconsole.txt"
_MS_NAME = 'HV4'


class TestMSEntryWrapper(unittest.TestCase):

    def setUp(self):
        super(TestMSEntryWrapper, self).setUp()

        self.ms_http = pvmhttp.load_pvm_resp(_MS_HTTPRESP_FILE)
        self.assertNotEqual(self.ms_http, None,
                            "Could not load %s " %
                            _MS_HTTPRESP_FILE)

        entries = self.ms_http.response.feed.findentries(
            ms._SYSTEM_NAME, _MS_NAME)

        self.assertNotEqual(entries, None,
                            "Could not find %s in %s" %
                            (_MS_NAME, _MS_HTTPRESP_FILE))

        self.myentry = entries[0]
        self.wrapper = ms.System.wrap(self.myentry)

        # Set a hardcoded value for MemoryUsedByHypervisor
        fw_mem = self.myentry.element.find(
            'AssociatedSystemMemoryConfiguration/MemoryUsedByHypervisor')
        fw_mem.text = '1536'

        mc_http = pvmhttp.load_pvm_resp(_MC_HTTPRESP_FILE)
        self.assertNotEqual(mc_http, None,
                            "Could not load %s" %
                            _MC_HTTPRESP_FILE)

        self.test_ioslot = self.wrapper.asio_config.io_slots[0]

        """ Create a bad wrapper to use when
            retrieving properties which don't exist """
        self.bad_wrapper = ms.System.wrap(mc_http.response.feed.entries[0])

    def verify_equal(self, method_name, returned_value, expected_value):
        self.assertEqual(returned_value, expected_value,
                         "%s returned %s instead of %s"
                         % (method_name, returned_value, expected_value))

    def call_simple_getter(self,
                           method_name,
                           expected_value,
                           expected_bad_value):

        # Use __getattribute__ to dynamically call the method
        value = self.wrapper.__getattribute__(method_name)
        if callable(value):
            value = value()
        self.verify_equal(method_name, value, expected_value)

        bad_value = self.bad_wrapper.__getattribute__(method_name)
        if callable(bad_value):
            bad_value = bad_value()
        self.verify_equal(method_name, bad_value, expected_bad_value)

    def test_get_val_str(self):
        expected_value = _MS_NAME
        value = self.wrapper._get_val_str(ms._SYSTEM_NAME)

        self.verify_equal("_get_val_str", value, expected_value)

        expected_value = None
        value = self.wrapper._get_val_str('BogusName')

        self.verify_equal(
            "_get_val_str for BogusName ", value, expected_value)

    def test_get_model(self):
        self.assertEqual(self.wrapper.mtms.model, "E4A")

    def test_get_type(self):
        self.assertEqual(self.wrapper.mtms.machine_type, "8203")

    def test_get_serial(self):
        self.assertEqual(self.wrapper.mtms.serial, "ACE0001")

    def test_get_mtms_str(self):
        self.assertEqual(self.wrapper.mtms.mtms_str, '8203-E4A*ACE0001')

    def test_highest_compat_mode(self):
        self.assertEqual(self.wrapper.highest_compat_mode(), 7)

    def test_proc_compat_modes(self):
        expected = ('default', 'POWER5', 'POWER6', 'POWER6_Enhanced',
                    'POWER6_Plus_Enhanced', 'POWER7')
        self.assertEqual(self.wrapper.proc_compat_modes, expected)

    def test_get_proc_units(self):
        self.call_simple_getter("proc_units", 500.0, 0)

    def test_get_min_proc_units(self):
        self.call_simple_getter("min_proc_units", 0.05, 0)

    def test_get_proc_units_configurable(self):
        self.call_simple_getter("proc_units_configurable", 500.0, 0)

    def test_get_proc_units_avail(self):
        self.call_simple_getter("proc_units_avail", 500.0, 0)

    def test_get_memory_total(self):
        self.call_simple_getter("memory_total", 5767168, 0)

    def test_get_memory_free(self):
        self.call_simple_getter("memory_free", 5242752, 0)

    def test_get_host_ip_address(self):
        self.call_simple_getter("host_ip_address", '127.0.0.1', None)

    def test_get_firmware_memory(self):
        self.call_simple_getter("firmware_memory", 1536, 0)

    def test_page_table_ratio(self):
        self.call_simple_getter("page_table_ratio", 7, 0)

    def test_get_system_name(self):
        self.wrapper.set_parm_value(ms._SYSTEM_NAME, 'XYZ')
        name = self.wrapper.system_name
        self.verify_equal("system_name", name, 'XYZ')
        self.wrapper.set_parm_value(ms._SYSTEM_NAME, 'ABC')
        name = self.wrapper.system_name
        self.verify_equal("system_name", name, 'ABC')

    def test_max_procs_per_aix_linux_lpar(self):
        self.call_simple_getter("max_procs_per_aix_linux_lpar", 32, 0)
        # Test setter
        self.wrapper.max_procs_per_aix_linux_lpar = 64
        self.call_simple_getter("max_procs_per_aix_linux_lpar", 64, 0)
        # Test fallback condition.  Should retrieve max_sys_procs_limit
        self.wrapper.max_procs_per_aix_linux_lpar = 0
        self.call_simple_getter("max_procs_per_aix_linux_lpar",
                                self.wrapper.max_sys_procs_limit, 0)

    def test_max_vcpus_per_aix_linux_lpar(self):
        self.call_simple_getter("max_vcpus_per_aix_linux_lpar", 30, 0)
        # Test setter
        self.wrapper.max_vcpus_per_aix_linux_lpar = 60
        self.call_simple_getter("max_vcpus_per_aix_linux_lpar", 60, 0)
        # Test fallback condition.  Should retrieve max_sys_vcpus_limit
        self.wrapper.max_vcpus_per_aix_linux_lpar = 0
        self.call_simple_getter("max_vcpus_per_aix_linux_lpar",
                                self.wrapper.max_sys_vcpus_limit, 0)

    def test_vios_links(self):
        self.call_simple_getter(
            "vios_links",
            ('https://9.1.2.3:12443/rest/api/uom/ManagedSystem/caae9209-25e5-'
             '35cd-a71a-ed55c03f294d/VirtualIOServer/32F3530F-ECA0-4EAA-A37E-'
             '4B792C21AF70',), ())

    def test_asio_wwpns(self):
        self.assertEqual(self.wrapper.asio_config.avail_wwpns, 65536)

    def test_asio_wwpn_prefix(self):
        self.assertEqual(self.wrapper.asio_config.wwpn_prefix,
                         '12379814471884843981')

    def test_ioslot_bus_grp_required(self):
        self.assertEqual(self.test_ioslot.bus_grp_required, False)

    def test_ioslot_description(self):
        self.assertEqual(self.test_ioslot.description, 'I/O Processor')

    def test_ioslot_feat_codes(self):
        self.assertEqual(self.test_ioslot.feat_codes, 0)

    def test_ioslot_pci_class(self):
        self.assertEqual(self.test_ioslot.pci_class, 512)

    def test_ioslot_pci_dev_id(self):
        self.assertEqual(self.test_ioslot.pci_dev_id, 4660)

    def test_ioslot_pci_subsys_dev_id(self):
        self.assertEqual(self.test_ioslot.pci_subsys_dev_id, 1202)

    def test_ioslot_pci_rev_id(self):
        self.assertEqual(self.test_ioslot.pci_rev_id, 0)

    def test_ioslot_pci_vendor_id(self):
        self.assertEqual(self.test_ioslot.pci_vendor_id, 4116)

    def test_ioslot_pci_subsys_vendor_id(self):
        self.assertEqual(self.test_ioslot.pci_subsys_vendor_id, 4116)

    def test_ioslot_drc_index(self):
        self.assertEqual(self.test_ioslot.drc_index,
                         553713681)

    def test_ioslot_drc_name(self):
        self.assertEqual(self.test_ioslot.drc_name,
                         'U5294.001.CEC1234-P01-C011')

    def test_get_aix_capable(self):
        self.call_simple_getter("_aix_capable", False, True)

    def test_get_capabilities(self):
        good_cap = {'active_lpar_mobility_capable': True,
                    'inactive_lpar_mobility_capable': True,
                    'ibmi_lpar_mobility_capable': True,
                    'custom_mac_addr_capable': True,
                    'ibmi_restrictedio_capable': True,
                    'ibmi_nativeio_capable': False,
                    'simplified_remote_restart_capable': False,
                    'aix_capable': False,
                    'ibmi_capable': True,
                    'linux_capable': False,
                    'shared_processor_pool_capable': True,
                    'active_memory_expansion_capable': True,
                    'dynamic_srr_capable': True,
                    'vnic_capable': True,
                    'vnic_failover_capable': True,
                    'disable_secure_boot_capable': False}
        bad_cap = {'active_lpar_mobility_capable': False,
                   'inactive_lpar_mobility_capable': False,
                   'ibmi_lpar_mobility_capable': False,
                   'custom_mac_addr_capable': True,
                   'ibmi_restrictedio_capable': False,
                   'ibmi_nativeio_capable': False,
                   'simplified_remote_restart_capable': False,
                   'aix_capable': True,
                   'ibmi_capable': False,
                   'linux_capable': True,
                   'shared_processor_pool_capable': False,
                   'active_memory_expansion_capable': False,
                   'dynamic_srr_capable': False,
                   'vnic_capable': False,
                   'vnic_failover_capable': False,
                   'disable_secure_boot_capable': False}
        self.call_simple_getter("get_capabilities", good_cap,
                                bad_cap)

    def test_session_is_master(self):
        self.assertTrue(self.wrapper.session_is_master)

    def test_migration_data(self):
        expected_data = {'active_lpar_mobility_capable': True,
                         'inactive_lpar_mobility_capable': True,
                         'ibmi_lpar_mobility_capable': True,
                         'custom_mac_addr_capable': True,
                         'ibmi_restrictedio_capable': True,
                         'ibmi_nativeio_capable': False,
                         'simplified_remote_restart_capable': False,
                         'aix_capable': False,
                         'ibmi_capable': True,
                         'linux_capable': False,
                         'shared_processor_pool_capable': True,
                         'active_memory_expansion_capable': True,
                         'max_migration_ops_supported': 9,
                         'active_migrations_supported': 0,
                         'inactive_migrations_supported': 5,
                         'preferred_active_migrations_supported': 0,
                         'preferred_inactive_migrations_supported': 5,
                         'active_migrations_in_progress': 0,
                         'inactive_migrations_in_progress': 0,
                         'proc_compat': 'default,POWER5,POWER6,POWER6_Enhanced'
                                        ',POWER6_Plus_Enhanced,POWER7',
                         'dynamic_srr_capable': True,
                         'vnic_capable': True,
                         'vnic_failover_capable': True,
                         'disable_secure_boot_capable': False}
        result_data = self.wrapper.migration_data
        self.assertEqual(result_data, expected_data,
                         'The returned data did not match expected values')


class TestMTMS(unittest.TestCase):
    def test_mtms(self):
        mtms = mtmwrap.MTMS.bld(None, '1234-567*ABCDEF0')
        self.assertEqual(mtms.machine_type, '1234')
        self.assertEqual(mtms.model, '567')
        self.assertEqual(mtms.serial, 'ABCDEF0')
        self.assertEqual(mtms.mtms_str, '1234-567*ABCDEF0')
        # Setters
        mtms.machine_type = '4321'
        self.assertEqual(mtms.machine_type, '4321')
        mtms.model = '765'
        self.assertEqual(mtms.model, '765')
        mtms.serial = '0FEDCBA'
        self.assertEqual(mtms.serial, '0FEDCBA')
        self.assertEqual(mtms.mtms_str, '4321-765*0FEDCBA')

if __name__ == "__main__":
    unittest.main()



# Copyright 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import unittest

import pypowervm.tests.test_utils.test_wrapper_abc as twrap
import pypowervm.wrappers.job as jwrap

CORRECT_CDATA_PARAMVAL = (
    '<ParameterValue xmlns="http://www.ibm.com/xmlns/systems/power/firmware/'
    'web/mc/2012_10/" xmlns:JobResponse="http://www.ibm.com/xmlns/systems/'
    'power/firmware/web/mc/2012_10/" xmlns:ns2="http://www.w3.org/XML/1998/'
    'namespace/k2" xmlns:ns3="http://www.w3.org/1999/xhtml" kb="CUR" '
    'kxe="false"><![CDATA[<XML_LIST><general><version>1</version></general>'
    '<reliableITL>true</reliableITL><deviceList><device><vendor>IBM</vendor>'
    '<deviceID>6005076802810B0FD00000000000049F042145</deviceID><itlList>'
    '<itl><Iwwpn>21000024FF409CD0</Iwwpn><Twwpn>500507680245CAC0</Twwpn><Lua>'
    '6000000000000</Lua></itl></itlList></device></deviceList></XML_LIST>]]>'
    '</ParameterValue>'
    "\n\n            ").encode("utf-8")

CORRECT_CDATA_CONTENT = (
    '<XML_LIST><general><version>1</version></general><reliableITL>true'
    '</reliableITL><deviceList><device><vendor>IBM</vendor><deviceID>'
    '6005076802810B0FD00000000000049F042145</deviceID><itlList><itl><Iwwpn>'
    '21000024FF409CD0</Iwwpn><Twwpn>500507680245CAC0</Twwpn><Lua>'
    '6000000000000</Lua></itl></itlList></device></deviceList></XML_LIST>')


class TestCDATA(twrap.TestWrapper):

    file = 'cdata.xml'
    wrapper_class_to_test = jwrap.Job

    """Verify CDATA segments survive going into and out of the Adapter."""
    def test_cdata_request(self):
        pval = self.dwrap.entry.element.find(
            'JobRequestInstance/JobParameters/JobParameter/ParameterValue')
        out = pval.toxmlstring()
        self.assertEqual(out, CORRECT_CDATA_PARAMVAL,
                         "CDATA was not preserved in JobRequest!\n%s" % out)

    def test_cdata_results(self):
        resdict = self.dwrap.get_job_results_as_dict()
        out = resdict['inputXML']
        self.assertEqual(out, CORRECT_CDATA_CONTENT,
                         "CDATA was not preserved in Results!\n%s" % out)

if __name__ == '__main__':
    unittest.main()



# Copyright 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import re
import unittest
import uuid

from lxml import etree
import mock
import six
import testtools

import pypowervm.adapter as apt
import pypowervm.entities as ent
import pypowervm.tests.test_fixtures as fx
from pypowervm.tests.test_utils import pvmhttp
from pypowervm.tests.test_utils import test_wrapper_abc as twrap
import pypowervm.utils.uuid as pvm_uuid
import pypowervm.wrappers.cluster as clust
import pypowervm.wrappers.entry_wrapper as ewrap
import pypowervm.wrappers.iocard as card
import pypowervm.wrappers.logical_partition as lpar
import pypowervm.wrappers.managed_system as ms
import pypowervm.wrappers.network as net
import pypowervm.wrappers.storage as stor
import pypowervm.wrappers.vios_file as vf
import pypowervm.wrappers.virtual_io_server as vios

NET_BRIDGE_FILE = 'fake_network_bridge.txt'
LPAR_FILE = 'lpar.txt'
VIOS_FILE = 'fake_vios_feed.txt'
VNETS_FILE = 'nbbr_virtual_network.txt'
SYS_VNIC_FILE = 'vnic_feed.txt'
SYS_SRIOV_FILE = 'sys_with_sriov.txt'


def _assert_clusters_equal(tc, cl1, cl2):
    tc.assertEqual(cl1.name, cl2.name)
    tc.assertEqual(cl1.repos_pv.name, cl2.repos_pv.name)
    tc.assertEqual(cl1.repos_pv.udid, cl2.repos_pv.udid)
    tc.assertEqual(cl1.nodes[0].hostname, cl2.nodes[0].hostname)


class SubWrapper(ewrap.Wrapper):
    schema_type = 'SubWrapper'
    _type_and_uuid = 'SubWrapper_TestClass'

    def __init__(self, **kwargs):
        class Txt(object):
            def __init__(self, val):
                self.text = val

        super(SubWrapper, self).__init__()
        self.data = dict((k, Txt(v)) for k, v in six.iteritems(kwargs))

    def _find(self, prop_name, use_find_all=False):
        try:
            return self.data[prop_name]
        except KeyError:
            return None


class TestElement(twrap.TestWrapper):
    file = NET_BRIDGE_FILE
    wrapper_class_to_test = net.NetBridge

    def test_child_poaching(self):
        """Creating an element with children of another existing element.

        Ensure the existing element remains intact.
        """
        # How many load groups did we start with?
        num_lg = len(self.dwrap.load_grps)
        # Create a new element with a load group from the NetBridge as a child.
        newel = ent.Element('foo', None,
                            children=[self.dwrap.load_grps[0].element])
        # Element creation did not poach the load group from the NetBridge.
        self.assertEqual(num_lg, len(self.dwrap.load_grps))

        # pypowervm.entities.Element.append (not to be confused with
        # etree.Element.append or list.append) should behave the same way.
        newel.append(self.dwrap.load_grps[1].element)
        # TODO(IBM): ...but it doesn't.  See comment in that method.
        # self.assertEqual(num_lg, len(self.dwrap.load_grps))

    @mock.patch('lxml.etree.tostring')
    def test_toxmlstring(self, mock_tostring):
        newel = ent.Element('foo', None)
        # No args
        self.assertEqual(mock_tostring.return_value, newel.toxmlstring())
        mock_tostring.assert_called_once_with(newel.element)
        # With kwargs
        mock_tostring.reset_mock()
        self.assertEqual(mock_tostring.return_value, newel.toxmlstring(
            pretty=False))
        mock_tostring.assert_called_once_with(newel.element)
        mock_tostring.reset_mock()
        self.assertEqual(mock_tostring.return_value,
                         newel.toxmlstring(pretty=True))
        mock_tostring.assert_called_once_with(newel.element, pretty_print=True)


class TestElementList(twrap.TestWrapper):
    file = SYS_SRIOV_FILE
    wrapper_class_to_test = ms.System

    def setUp(self):
        super(TestElementList, self).setUp()
        self.pport = self.dwrap.asio_config.sriov_adapters[0].phys_ports[0]
        self.tag = 'ConfiguredOptions'

    def _validate_xml(self, val_list):
        outer_tag = self.pport.schema_type
        tag_before = 'ConfiguredMTU'
        tag_after = 'ConfiguredPortSwitchMode'
        # Opening
        tag_pat_fmt = r'<%s(\s[^>]*)?>'
        elem_pat_fmt = tag_pat_fmt + r'%s</%s>\s*'
        pattern = '.*'
        pattern += tag_pat_fmt % outer_tag
        pattern += '.*'
        pattern += elem_pat_fmt % (tag_before, '[^<]*', tag_before)
        for val in val_list:
            pattern += elem_pat_fmt % (self.tag, val, self.tag)
        pattern += elem_pat_fmt % (tag_after, '[^<]*', tag_after)
        pattern += '.*'
        pattern += tag_pat_fmt % ('/' + outer_tag)
        pattern += '.*'
        self.assertTrue(re.match(pattern.encode('utf-8'),
                                 self.pport.toxmlstring(), flags=re.DOTALL))

    def test_everything(self):
        """Ensure ElementList behaves like a list where implemented."""
        # Wrapper._get_elem_list, ElementList.__init__
        coel = self.pport._get_elem_list(self.tag)
        # index
        self.assertEqual(0, coel.index('autoDuplex'))
        self.assertEqual(1, coel.index('Veb'))
        self.assertRaises(ValueError, coel.index, 'foo')
        # __len__
        self.assertEqual(2, len(coel))
        # __repr__
        self.assertEqual("['autoDuplex', 'Veb']", repr(coel))
        # __contains__
        self.assertIn('autoDuplex', coel)
        self.assertIn('Veb', coel)
        self.assertNotIn('foo', coel)
        # __str__
        self.assertEqual("['autoDuplex', 'Veb']", str(coel))
        # __getitem__
        self.assertEqual('autoDuplex', coel[0])
        self.assertEqual('Veb', coel[1])
        self.assertRaises(IndexError, coel.__getitem__, 2)
        # __setitem__
        coel[0] = 'fullDuplex'
        self.assertEqual('fullDuplex', coel[0])
        self.assertRaises(IndexError, coel.__setitem__, 2, 'foo')
        # append
        coel.append('foo')
        self._validate_xml(['fullDuplex', 'Veb', 'foo'])
        # extend
        coel.extend(['bar', 'baz'])
        self._validate_xml(['fullDuplex', 'Veb', 'foo', 'bar', 'baz'])
        # __delitem__
        del coel[3]
        self._validate_xml(['fullDuplex', 'Veb', 'foo', 'baz'])
        # remove
        coel.remove('foo')
        self._validate_xml(['fullDuplex', 'Veb', 'baz'])
        # __iter__
        self.assertEqual(['fullDuplex', 'Veb', 'baz'], [val for val in coel])
        # clear
        coel.clear()
        self.assertEqual(0, len(coel))
        self._validate_xml([])
        # Inserting stuff back in puts it in the right place
        coel.extend(['one', 'two', 'three'])
        self._validate_xml(['one', 'two', 'three'])
        # Wrapper._set_elem_list
        self.pport._set_elem_list(self.tag, ['four', 'five', 'six'])
        self._validate_xml(['four', 'five', 'six'])


class TestWrapper(unittest.TestCase):
    def test_get_val_str(self):
        w = SubWrapper(one='1', foo='foo', empty='')
        self.assertEqual(w._get_val_str('one'), '1')
        self.assertEqual(w._get_val_str('foo'), 'foo')
        self.assertEqual(w._get_val_str('empty'), '')
        self.assertIsNone(w._get_val_str('nonexistent'))
        self.assertEqual(w._get_val_str('nonexistent', default='10'), '10')

    def test_get_val_percent(self):
        w = SubWrapper(one='2.45%', two='2.45', three=None, four='123',
                       five='1.2345', six='123.0', seven='123%',
                       eight='%123', nine='12%3')
        self.assertEqual(w._get_val_percent('one'), 0.0245)
        self.assertEqual(w._get_val_percent('two'), 0.0245)
        self.assertEqual(w._get_val_percent('three'), None)
        self.assertEqual(w._get_val_percent('four'), 1.23)
        self.assertEqual(w._get_val_percent('five'), 0.012345)
        self.assertEqual(w._get_val_percent('six'), 1.23)
        self.assertEqual(w._get_val_percent('seven'), 1.23)
        self.assertEqual(w._get_val_percent('eight'), 1.23)
        # Interesting test:
        self.assertEqual(w._get_val_percent('nine'), 0.12)
        self.assertIsNone(w._get_val_percent('nonexistent'))

    def test_get_val_int(self):
        w = SubWrapper(one='1', nan='foo', empty='')
        self.assertEqual(w._get_val_int('one'), 1)
        self.assertIsNone(w._get_val_int('nan'))
        self.assertIsNone(w._get_val_int('empty'))
        self.assertIsNone(w._get_val_int('nonexistent'))
        self.assertEqual(w._get_val_int('nonexistent', default=10), 10)

    def test_get_val_float(self):
        w = SubWrapper(one='1', two_point_two='2.2', nan='foo', empty='')
        self.assertAlmostEqual(w._get_val_float('one'), 1)
        self.assertAlmostEqual(w._get_val_float('two_point_two'), 2.2)
        self.assertIsNone(w._get_val_float('nan'))
        self.assertIsNone(w._get_val_float('empty'))
        self.assertIsNone(w._get_val_float('nonexistent'))
        self.assertAlmostEqual(w._get_val_float('one', default=2), 1)
        self.assertAlmostEqual(w._get_val_float('two_point_two', default=3),
                               2.2)
        self.assertAlmostEqual(w._get_val_float('nan', default=1), 1)
        self.assertAlmostEqual(w._get_val_float('empty', default=1), 1)
        self.assertAlmostEqual(w._get_val_int('nonexistent', default=1.7), 1.7)

    def test_get_val_bool(self):
        w = SubWrapper(one='1', t='true', T='TRUE', f='false', F='FALSE',
                       empty='')
        self.assertTrue(w._get_val_bool('t'))
        self.assertTrue(w._get_val_bool('T'))
        self.assertFalse(w._get_val_bool('one'))
        self.assertFalse(w._get_val_bool('empty'))
        self.assertFalse(w._get_val_bool('f'))
        self.assertFalse(w._get_val_bool('F'))
        self.assertFalse(w._get_val_bool('nonexistent'))
        self.assertTrue(w._get_val_bool('t', default=False))
        self.assertTrue(w._get_val_bool('T', default=False))
        self.assertFalse(w._get_val_bool('one', default=True))
        self.assertFalse(w._get_val_bool('empty', default=True))
        self.assertFalse(w._get_val_bool('f', default=True))
        self.assertFalse(w._get_val_bool('F', default=True))
        self.assertTrue(w._get_val_bool('nonexistent', default=True))


class TestEntryWrapper(testtools.TestCase):

    def setUp(self):
        super(TestEntryWrapper, self).setUp()
        self.adpt = self.useFixture(fx.AdapterFx()).adpt

    def test_etag(self):
        fake_entry = ent.Entry({}, ent.Element('fake_entry', self.adpt),
                               self.adpt)
        etag = '1234'
        ew = ewrap.EntryWrapper.wrap(fake_entry, etag=etag)
        self.assertEqual(etag, ew.etag)

        ew = ewrap.EntryWrapper.wrap(fake_entry)
        self.assertEqual(None, ew.etag)

    def test_set_uuid(self):

        # Test that an AttributeError is raised
        def set_wrap_uuid(wrap, value):
            wrap.uuid = value
        self.assertRaises(AttributeError, set_wrap_uuid,
                          ewrap.EntryWrapper(None), 'fake-uuid-value')

        # Test that we call the mixin set_uuid method for valid cases.
        class ValidEntryWrap(ewrap.EntryWrapper, ewrap.WrapperSetUUIDMixin):
            pass
        with mock.patch('pypowervm.wrappers.entry_wrapper.WrapperSetUUIDMixin'
                        '.set_uuid') as mock_setup:
            uuid1 = pvm_uuid.convert_uuid_to_pvm(str(uuid.uuid4()))
            ValidEntryWrap(None).uuid = uuid1
            mock_setup.assert_called_with(uuid1)

    def test_load(self):
        etag = '1234'
        resp = apt.Response('reqmethod', 'reqpath', 'status', 'reason',
                            dict(etag=etag))

        # Entry or Feed is not set, so expect an exception
        self.assertRaises(KeyError, ewrap.EntryWrapper.wrap, resp)

        # Set an entry...
        entry = ent.Entry({}, ent.Element('entry', self.adpt), self.adpt)
        resp.entry = entry

        # Run
        ew = ewrap.EntryWrapper.wrap(resp)

        # Validate
        self.assertEqual(entry, ew.entry)
        self.assertEqual(etag, ew.etag)

        # Create a response with no headers
        resp2 = apt.Response('reqmethod', 'reqpath', 'status', 'reason', {})
        resp2.entry = entry
        # Run
        ew = ewrap.EntryWrapper.wrap(resp2)
        # Validate the etag is None since there were no headers
        self.assertEqual(None, ew.etag)

        # Wipe our entry, add feed.
        resp.entry = None
        e1 = ent.Entry({'etag': '1'}, ent.Element('e1', self.adpt),
                       self.adpt)
        e2 = ent.Entry({'etag': '2'}, ent.Element('e2', self.adpt),
                       self.adpt)
        resp.feed = ent.Feed({}, [e1, e2])

        # Run
        ew = ewrap.EntryWrapper.wrap(resp)

        # Validate
        self.assertEqual(e1, ew[0].entry)
        self.assertEqual('1', ew[0].etag)
        self.assertEqual(e2, ew[1].entry)
        self.assertEqual('2', ew[1].etag)

    @mock.patch('lxml.etree.tostring')
    def test_toxmlstring(self, mock_tostring):
        wrp = ewrap.EntryWrapper.wrap(ent.Entry(
            {}, ent.Element('fake_entry', None), None))
        # No args
        self.assertEqual(mock_tostring.return_value, wrp.toxmlstring())
        mock_tostring.assert_called_once_with(wrp.entry.element)
        # With kwargs
        mock_tostring.reset_mock()
        self.assertEqual(mock_tostring.return_value, wrp.toxmlstring(
            pretty=False))
        mock_tostring.assert_called_once_with(wrp.entry.element)
        mock_tostring.reset_mock()
        self.assertEqual(mock_tostring.return_value, wrp.toxmlstring(
            pretty=True))
        mock_tostring.assert_called_once_with(
            wrp.entry.element, pretty_print=True)


class TestElementWrapper(testtools.TestCase):
    """Tests for the ElementWrapper class."""

    def setUp(self):
        super(TestElementWrapper, self).setUp()
        self.adpt = self.useFixture(fx.AdapterFx()).adpt
        self.resp = pvmhttp.load_pvm_resp(NET_BRIDGE_FILE).get_response()
        self.nb1 = ewrap.EntryWrapper.wrap(self.resp.feed.entries[0])
        self.resp2 = pvmhttp.load_pvm_resp(NET_BRIDGE_FILE).get_response()
        self.nb2 = ewrap.EntryWrapper.wrap(self.resp2.feed.entries[0])

    def test_equality(self):
        """Validates that two elements loaded from the same data is equal."""
        sea1 = self._find_seas(self.nb1.entry)[0]
        sea2 = self._find_seas(self.nb2.entry)[0]
        sea2copy = copy.deepcopy(sea2)
        self.assertTrue(sea1 == sea2)
        self.assertEqual(sea2, sea2copy)

        # Change the other SEA
        sea2.element.element.append(etree.Element('Bob'))
        self.assertFalse(sea1 == sea2)

    def test_inequality_by_subelem_change(self):
        sea1 = self._find_seas(self.nb1.entry)[0]
        sea2 = self._find_seas(self.nb2.entry)[0]
        sea_trunk = sea2.element.findall('TrunkAdapters/TrunkAdapter')[0]
        pvid = sea_trunk.find('PortVLANID')
        pvid.text = '1'
        self.assertFalse(sea1 == sea2)

    def test_unequal(self):
        sea1 = self._find_seas(self.nb1.entry)[0]
        sea2 = self._find_seas(self.nb2.entry)[0]
        self.assertEqual(sea1, sea2)
        # Different text makes 'em different
        sea1.element.text = 'Bogus'
        self.assertNotEqual(sea1, sea2)
        # reset
        sea1.element.text = sea2.element.element.text
        # Different tag makes 'em different
        sea1.element.tag = 'Bogus'
        self.assertNotEqual(sea1, sea2)

    def _find_seas(self, entry):
        """Wrapper for the SEAs."""
        found = entry.element.find('SharedEthernetAdapters')
        return ewrap.WrapperElemList(found, net.SEA)

    def test_fresh_element(self):
        # Default: UOM namespace, no <Metadata/>
        class MyElement(ewrap.ElementWrapper):
            schema_type = 'SomePowerObject'
        myel = MyElement._bld(self.adpt)
        self.assertEqual(myel.schema_type, 'SomePowerObject')
        self.assertEqual(myel.element.toxmlstring(),
                         '<SomePowerObject/>'.encode("utf-8"))

        # Can override namespace and attrs and trigger inclusion of <Metadata/>
        class MyElement3(ewrap.ElementWrapper):
            schema_type = 'SomePowerObject'
            default_attrib = {'foo': 'bar'}
            schema_ns = 'baz'
            has_metadata = True
        myel = MyElement3._bld(self.adpt)
        self.assertEqual(
            myel.element.toxmlstring(),
            '<ns0:SomePowerObject xmlns:ns0="baz" foo="bar"><ns0:Metadata>'
            '<ns0:Atom/></ns0:Metadata></ns0:SomePowerObject>'.encode("utf-8"))

        # Same thing, but via the decorator
        @ewrap.ElementWrapper.pvm_type('SomePowerObject', has_metadata=True,
                                       ns='baz', attrib={'foo': 'bar'})
        class MyElement4(ewrap.ElementWrapper):
            pass
        myel = MyElement4._bld(self.adpt)
        self.assertEqual(
            myel.element.toxmlstring(),
            '<ns0:SomePowerObject xmlns:ns0="baz" foo="bar"><ns0:Metadata>'
            '<ns0:Atom/></ns0:Metadata></ns0:SomePowerObject>'.encode("utf-8"))

        # Now 'SomePowerObject' is registered.  Prove that we can use wrap() to
        # instantiate MyElement4 straight from ElementWrapper.
        el = ent.Element('SomePowerObject', self.adpt, ns='baz',
                         attrib={'foo': 'bar'})
        w = ewrap.ElementWrapper.wrap(el)
        self.assertIsInstance(w, MyElement4)

    def test_href(self):
        path = 'LoadGroups/LoadGroup/VirtualNetworks/link'
        # Get all
        hrefs = self.nb1.get_href(path)
        self.assertEqual(len(hrefs), 13)
        self.assertEqual(
            hrefs[2],
            'https://9.1.2.3:12443/rest/api/uom/ManagedSystem/726e9cb3-6576-3d'
            'f5-ab60-40893d51d074/VirtualNetwork/2b4ab8ea-4b15-3430-b2cd-45954'
            'cfaba0d')
        # Request one - should return None
        hrefs = self.nb1.get_href(path, one_result=True)
        self.assertIsNone(hrefs)
        # set_href should refuse to set multiple links
        self.assertRaises(ValueError, self.nb1.set_href, path, 'foo')

        # Drill down to the (only) SEA
        sea = ewrap.ElementWrapper.wrap(
            self.nb1._find('SharedEthernetAdapters/SharedEthernetAdapter'))
        path = 'AssignedVirtualIOServer'
        hrefs = sea.get_href(path)
        self.assertEqual(len(hrefs), 1)
        self.assertEqual(
            hrefs[0],
            'https://9.1.2.3:12443/rest/api/uom/ManagedSystem/726e9cb3-6576-3d'
            'f5-ab60-40893d51d074/VirtualIOServer/691019AF-506A-4896-AADE-607E'
            '21FA93EE')
        # Now make sure one_result returns the string (not a list)
        href = sea.get_href(path, one_result=True)
        self.assertEqual(
            href,
            'https://9.1.2.3:12443/rest/api/uom/ManagedSystem/726e9cb3-6576-3d'
            'f5-ab60-40893d51d074/VirtualIOServer/691019AF-506A-4896-AADE-607E'
            '21FA93EE')
        # Test setter
        sea.set_href(path, 'foo')
        self.assertEqual(sea.get_href(path, one_result=True), 'foo')
        # Now try setting one that doesn't exist.  First on a top-level path.
        path = 'NewElement'
        sea.set_href(path, 'bar')
        self.assertEqual(sea.get_href(path, one_result=True), 'bar')
        # ...and now on a nested path.
        path = 'BackingDeviceChoice/EthernetBackingDevice/NewLink'
        sea.set_href(path, 'baz')
        self.assertEqual(sea.get_href(path, one_result=True), 'baz')

    def _verify_element_clone(self, el1, el2):
        # Equal according to _element_equality
        self.assertEqual(el1, el2)
        # Not the same reference
        self.assertIsNot(el1, el2)
        # Adapter references are the same
        self.assertIs(el1.adapter, el2.adapter)
        # etree.Elements are not the same reference
        self.assertIsNot(el1.element, el2.element)
        # But they marshal identically
        self.assertEqual(el1.toxmlstring().strip(),
                         el2.toxmlstring().strip())

    def test_element_clone(self):
        el1 = self.nb1.element
        el2 = copy.deepcopy(el1)
        self._verify_element_clone(el1, el2)

    def _verify_properties_clone(self, props1, props2):
        # Properties should be deeply equal
        self.assertEqual(props1, props2)
        # But not the same reference
        self.assertIsNot(props1, props2)
        # Strings are shared copy-on-write.  Ensure changing one does not
        # change the other
        props1['id'] = 'abc'
        self.assertNotEqual(props1['id'], props2['id'])

    def _verify_entry_clone(self, ent1, ent2):
        # Elements should be cloned to the same spec as Element.__deepcopy__()
        self._verify_element_clone(ent1.element, ent2.element)
        self._verify_properties_clone(ent1.properties, ent2.properties)
        # Ensure deep copy - sub-properties also not the same reference.
        links1 = ent1.properties['links']
        links2 = ent2.properties['links']
        self.assertIsNot(links1, links2)
        # And one more layer down
        self.assertIsNot(links1['SELF'], links2['SELF'])

    def test_entry_clone(self):
        ent1 = self.nb1.entry
        ent2 = copy.deepcopy(ent1)
        self._verify_entry_clone(ent1, ent2)

    def _verify_feed_clone(self, feed1, feed2):
        self._verify_properties_clone(feed1.properties, feed2.properties)
        self.assertEqual(len(feed1.entries), len(feed2.entries))
        for ent1, ent2 in zip(feed1.entries, feed2.entries):
            self._verify_entry_clone(ent1, ent2)

    def test_feed_clone(self):
        feed1 = self.resp.feed
        feed2 = copy.deepcopy(feed1)
        self._verify_feed_clone(feed1, feed2)

    def _verify_response_clone(self, resp1, resp2):
        for attr in ('reqmethod', 'reqpath', 'reqheaders', 'reqbody', 'status',
                     'reason', 'headers', 'body', 'adapter'):
            self.assertEqual(getattr(resp1, attr), getattr(resp2, attr))
        self.assertIsNot(resp1.headers, resp2.headers)
        self.assertIs(resp1.adapter, resp2.adapter)
        if resp1.feed is None:
            self.assertIsNone(resp2.feed)
        else:
            self._verify_feed_clone(resp1.feed, resp2.feed)
        if resp1.entry is None:
            self.assertIsNone(resp2.entry)
        else:
            self._verify_entry_clone(resp1.entry, resp2.entry)

    def test_response_clone(self):
        # Network Bridge Response has a feed
        resp1 = self.resp
        resp2 = copy.deepcopy(resp1)
        self._verify_response_clone(resp1, resp2)
        # This one has entry
        resp3 = pvmhttp.load_pvm_resp(
            'get_volume_group_no_rep.txt').get_response()
        resp4 = copy.deepcopy(resp3)
        self._verify_response_clone(resp3, resp4)

    def test_entrywrapper_clone(self):
        ew1 = self.nb1
        ew2 = copy.deepcopy(ew1)
        # Entrys should be cloned to the same spec as Entry.__deepcopy__()
        self._verify_entry_clone(ew1.entry, ew2.entry)
        # Etags should match
        self.assertEqual(ew1.etag, ew2.etag)
        # But changing one should not change the other
        ew1._etag = 'abc'
        self.assertNotEqual(ew1.etag, ew2.etag)

    def test_elementwrapper_clone(self):
        ew1 = self.nb1.seas[0]
        ew2 = copy.deepcopy(ew1)
        # Elements should be cloned to the same spec as Element.__deepcopy__()
        self._verify_element_clone(ew1.element, ew2.element)


class TestWrapperElemList(testtools.TestCase):
    """Tests for the WrapperElemList class."""

    def setUp(self):
        super(TestWrapperElemList, self).setUp()
        self.adpt = self.useFixture(fx.AdapterFx()).adpt
        # No indirect
        self.seas_wel = net.NetBridge.wrap(pvmhttp.load_pvm_resp(
            NET_BRIDGE_FILE).get_response())[0].seas
        # With indirect
        self.backdev_wel = card.VNIC.wrap(pvmhttp.load_pvm_resp(
            SYS_VNIC_FILE).get_response())[0].back_devs

    def test_get(self):
        self.assertIsInstance(self.seas_wel[0], net.SEA)
        self.assertRaises(IndexError, lambda a, i: a[i], self.seas_wel, 2)
        # Works with indirect
        self.assertIsInstance(self.backdev_wel[0], card.VNICBackDev)
        self.assertRaises(IndexError, lambda a, i: a[i], self.backdev_wel, 2)

    def test_length(self):
        self.assertEqual(2, len(self.seas_wel))
        self.assertEqual(2, len(self.backdev_wel))

    def test_append(self):
        sea_add = ewrap.ElementWrapper.wrap(
            ent.Element('SharedEthernetAdapter', self.adpt))
        self.assertEqual(2, len(self.seas_wel))

        # Test Append
        self.seas_wel.append(sea_add)
        self.assertEqual(3, len(self.seas_wel))
        # Appending to indirect
        backdev = copy.deepcopy(self.backdev_wel[0])
        self.backdev_wel.append(backdev)
        self.assertEqual(3, len(self.backdev_wel))

        # Make sure we can also remove what was just added.
        self.seas_wel.remove(sea_add)
        self.assertEqual(2, len(self.seas_wel))
        # Removing from indirect
        self.backdev_wel.remove(backdev)
        self.assertEqual(2, len(self.backdev_wel))

    def test_extend(self):
        seas = [
            ewrap.ElementWrapper.wrap(ent.Element('SharedEthernetAdapter',
                                                  self.adpt)),
            ewrap.ElementWrapper.wrap(ent.Element('SharedEthernetAdapter',
                                                  self.adpt))
        ]
        self.assertEqual(2, len(self.seas_wel))
        self.seas_wel.extend(seas)
        self.assertEqual(4, len(self.seas_wel))
        self.adpt.build_href.return_value = 'href'
        # Extending indirect
        backdevs = [card.VNICBackDev.bld(self.adpt, 'vios_uuid', 1, 2),
                    card.VNICBackDev.bld(self.adpt, 'vios_uuid', 3, 4)]
        self.backdev_wel.extend(backdevs)
        self.assertEqual(4, len(self.backdev_wel))

        # Make sure that we can also remove what we added.  We remove a
        # logically identical element to test the equivalence function
        e = ewrap.ElementWrapper.wrap(ent.Element('SharedEthernetAdapter',
                                                  self.adpt))
        self.seas_wel.remove(e)
        self.seas_wel.remove(e)
        self.assertEqual(2, len(self.seas_wel))
        # With indirect
        self.backdev_wel.remove(card.VNICBackDev.bld(self.adpt, 'vios_uuid', 1,
                                                     2))
        self.assertEqual(3, len(self.backdev_wel))
        # Non-equivalent one doesn't work
        self.assertRaises(ValueError, self.backdev_wel.remove,
                          card.VNICBackDev.bld(self.adpt, 'vios_uuid', 1, 3))

    def test_in(self):
        # This really does fail without __contains__
        self.assertIn(self.seas_wel[0], self.seas_wel)
        # Works with indirect
        self.assertIn(self.backdev_wel[0], self.backdev_wel)

    def test_index(self):
        self.assertEqual(self.seas_wel.index(self.seas_wel[0]), 0)
        # Works with indirect
        self.assertEqual(self.backdev_wel.index(self.backdev_wel[0]), 0)

    def test_str(self):
        strout = str(self.seas_wel)
        self.assertEqual('[', strout[0])
        self.assertEqual(']', strout[-1])
        for chunk in strout.split(','):
            self.assertIn('SEA', chunk)
        # And for indirect
        strout = str(self.backdev_wel)
        self.assertEqual('[', strout[0])
        self.assertEqual(']', strout[-1])
        for chunk in strout.split(','):
            self.assertIn('VNIC', chunk)

    def test_repr(self):
        strout = repr(self.seas_wel)
        self.assertEqual('[', strout[0])
        self.assertEqual(']', strout[-1])
        for chunk in strout.split(','):
            self.assertIn('SEA', chunk)
        # And for indirect
        strout = repr(self.backdev_wel)
        self.assertEqual('[', strout[0])
        self.assertEqual(']', strout[-1])
        for chunk in strout.split(','):
            self.assertIn('VNIC', chunk)


class TestActionableList(unittest.TestCase):
    """Tests for the Actionable List class."""

    def test_extend(self):
        def test(new_list):
            self.assertEqual([1, 2, 3, 4, 5], new_list)
        l = ewrap.ActionableList([1, 2, 3], test)

        # Extend here.
        l.extend([4, 5])
        self.assertEqual(5, len(l))
        self.assertEqual(5, l[4])

    def test_append(self):
        def test(new_list):
            self.assertEqual([1, 2, 3, 4], new_list)
        l = ewrap.ActionableList([1, 2, 3], test)

        # Append here.
        l.append(4)
        self.assertEqual(4, len(l))
        self.assertEqual(4, l[3])

    def test_remove(self):
        def test(new_list):
            self.assertEqual([1, 3], new_list)
        l = ewrap.ActionableList([1, 2, 3], test)

        # Remove here.
        l.remove(2)
        self.assertEqual(2, len(l))
        self.assertEqual(3, l[1])

    def test_insert(self):
        def test(new_list):
            self.assertEqual([1, 2, 3, 4], new_list)
        l = ewrap.ActionableList([1, 2, 3], test)

        # Insert here.
        l.insert(3, 4)
        self.assertEqual(4, len(l))
        self.assertEqual(4, l[3])

    def test_pop(self):
        def test(new_list):
            self.assertEqual([1, 2], new_list)
        l = ewrap.ActionableList([1, 2, 3], test)

        # Pop here.
        l.pop(2)
        self.assertEqual(2, len(l))
        self.assertEqual(2, l[1])

    def test_complex_path(self):
        function = mock.MagicMock()

        l = ewrap.ActionableList([1, 2, 3], function)
        self.assertEqual(3, len(l))
        self.assertEqual(3, l[2])

        # Try extending
        l.extend([4, 5])
        self.assertEqual(5, len(l))
        self.assertEqual(5, l[4])

        # Try appending
        l.append(6)
        self.assertEqual(6, len(l))
        self.assertEqual(6, l[5])

        # Try removing
        l.remove(6)
        self.assertEqual(5, len(l))
        self.assertEqual(5, l[4])

        # Try inserting
        l.insert(5, 6)
        self.assertEqual(6, len(l))
        self.assertEqual(6, l[5])

        # Try popping
        self.assertEqual(6, l.pop(5))
        self.assertEqual(5, len(l))
        self.assertEqual(5, l[4])

        # Make sure our function was called each time
        self.assertEqual(5, function.call_count)


class TestGet(testtools.TestCase):
    """Tests for EntryWrapper.get()."""

    def setUp(self):
        super(TestGet, self).setUp()
        self.adpt = self.useFixture(fx.AdapterFx()).adpt

    @mock.patch('pypowervm.wrappers.logical_partition.LPAR.wrap')
    def test_get_root(self, mock_wrap):
        """Various permutations of EntryWrapper.get on a ROOT object."""
        # Happy path - feed.  Ensure misc args are passed through.
        lpar.LPAR.get(self.adpt, foo='bar', baz=123)
        self.adpt.read.assert_called_with(lpar.LPAR.schema_type, foo='bar',
                                          baz=123)
        mock_wrap.assert_called_with(self.adpt.read.return_value)
        mock_wrap.reset_mock()
        # Happy path - entry with 'uuid'
        lpar.LPAR.get(self.adpt, uuid='123')
        self.adpt.read.assert_called_with(lpar.LPAR.schema_type, root_id='123')
        mock_wrap.assert_called_with(self.adpt.read.return_value)
        mock_wrap.reset_mock()
        # Happy path - entry with 'root_id'
        lpar.LPAR.get(self.adpt, root_id='123')
        self.adpt.read.assert_called_with(lpar.LPAR.schema_type, root_id='123')
        mock_wrap.assert_called_with(self.adpt.read.return_value)
        mock_wrap.reset_mock()

    @mock.patch('pypowervm.wrappers.network.CNA.wrap')
    def test_get_child(self, mock_wrap):
        """Various permutations of EntryWrapper.get on a CHILD object."""
        # Happy path - feed.  Parent specified as class
        net.CNA.get(self.adpt, parent_type=lpar.LPAR, parent_uuid='123')
        self.adpt.read.assert_called_with(lpar.LPAR.schema_type, root_id='123',
                                          child_type=net.CNA.schema_type)
        mock_wrap.assert_called_with(self.adpt.read.return_value)
        mock_wrap.reset_mock()
        # Happy path - entry with 'uuid'.  Parent specified as string.
        net.CNA.get(
            self.adpt, parent_type=lpar.LPAR.schema_type, parent_uuid='123',
            uuid='456')
        self.adpt.read.assert_called_with(
            lpar.LPAR.schema_type, root_id='123',
            child_type=net.CNA.schema_type, child_id='456')
        mock_wrap.assert_called_with(self.adpt.read.return_value)
        mock_wrap.reset_mock()
        # Happy path - entry with 'child_id'.  Parent specified as instance.
        parent = mock.Mock(spec=lpar.LPAR, schema_type=lpar.LPAR.schema_type,
                           uuid='123')
        net.CNA.get(self.adpt, parent=parent, child_id='456')
        self.adpt.read.assert_called_with(
            lpar.LPAR.schema_type, root_id='123',
            child_type=net.CNA.schema_type, child_id='456')
        mock_wrap.assert_called_with(self.adpt.read.return_value)
        mock_wrap.reset_mock()

    @mock.patch('pypowervm.wrappers.entry_wrapper.EntryWrapper.wrap')
    def test_get_errors(self, mock_wrap):
        """Error paths in EntryWrapper.get."""
        # parent_type specified, parent_uuid not.
        self.assertRaises(ValueError,
                          net.CNA.get, self.adpt, parent_type=lpar.LPAR)
        # CHILD mode forbids 'root_id' (must use 'parent_uuid').
        self.assertRaises(ValueError, net.CNA.get, self.adpt,
                          parent_type=lpar.LPAR, parent_uuid='1', root_id='2')
        # CHILD mode can't have both 'uuid' and 'child_id'.
        self.assertRaises(ValueError, net.CNA.get, self.adpt,
                          parent_type=lpar.LPAR, parent_uuid='12', uuid='34',
                          child_id='56')
        # ROOT mode forbids parent_uuid.
        self.assertRaises(ValueError,
                          lpar.LPAR.get, self.adpt, parent_uuid='123')
        # ROOT mode forbids child_type.
        self.assertRaises(ValueError,
                          lpar.LPAR.get, self.adpt, child_type=net.CNA)
        # ROOT mode forbids child_id.
        self.assertRaises(ValueError,
                          lpar.LPAR.get, self.adpt, child_id='123')
        # ROOT mode can't have both 'uuid' and 'root_id'.
        self.assertRaises(ValueError,
                          lpar.LPAR.get, self.adpt, uuid='12', root_id='34')
        # Nothing was ever wrapped
        mock_wrap.assert_not_called()

    @mock.patch('pypowervm.wrappers.entry_wrapper.EntryWrapper.wrap')
    def test_get_by_href(self, mock_wrap):
        self.assertEqual(
            mock_wrap.return_value,
            ewrap.EntryWrapper.get_by_href(self.adpt, 'href', one=2, three=4))
        self.adpt.read_by_href.assert_called_once_with('href', one=2, three=4)


class TestSearch(testtools.TestCase):
    """Tests for EntryWrapper.search()."""

    def setUp(self):
        super(TestSearch, self).setUp()
        self.adp = apt.Adapter(self.useFixture(fx.SessionFx()).sess)

    def _validate_request(self, path, *feedcontents):
        def validate_request(meth, _path, *args, **kwargs):
            self.assertTrue(_path.endswith(path))
            resp = apt.Response('meth', 'path', 'status', 'reason', {},
                                reqheaders={'Accept': ''})
            resp.feed = ent.Feed({}, feedcontents)
            return resp
        return validate_request

    @mock.patch('pypowervm.adapter.Adapter._request')
    def test_good(self, mock_rq):
        def validate_result(clwrap):
            self.assertIsInstance(clwrap, clust.Cluster)
            self.assertEqual(clwrap.name, 'cl1')
            self.assertEqual(clwrap.repos_pv.name, 'hdisk1')
            self.assertEqual(clwrap.nodes[0].hostname, 'vios1')

        mock_rq.side_effect = self._validate_request(
            "/rest/api/uom/Cluster/search/(ClusterName=='cl1')?group=None",
            clust.Cluster.bld(self.adp, 'cl1', stor.PV.bld(self.adp, 'hdisk1',
                                                           'udid1'),
                              clust.Node.bld(
                                  self.adp, hostname='vios1')).entry)

        clwraps = clust.Cluster.search(self.adp, name='cl1')
        self.assertEqual(len(clwraps), 1)
        validate_result(clwraps[0])
        # Test one_result on a registered key with a single hit
        validate_result(clust.Cluster.search(self.adp, one_result=True,
                                             name='cl1'))

    @mock.patch('pypowervm.adapter.Adapter._request')
    def test_negate(self, mock_rq):
        mock_rq.side_effect = self._validate_request(
            "/rest/api/uom/Cluster/search/(ClusterName!='cl1')?group=None")
        clwraps = clust.Cluster.search(self.adp, negate=True, name='cl1')
        self.assertEqual(clwraps, [])
        # Test one_result with no hits
        self.assertIsNone(clust.Cluster.search(self.adp, negate=True,
                                               one_result=True, name='cl1'))

    def test_no_such_search_key(self):
        """Ensure an invalid search key gives ValueError."""
        self.assertRaises(ValueError, clust.Cluster.search, self.adp,
                          foo='bar')

    @mock.patch('pypowervm.adapter.Adapter._request')
    def test_quote(self, mock_rq):
        """Ensure special chars in the search value are properly encoded."""
        mock_rq.side_effect = self._validate_request(
            "/rest/api/uom/Cluster/search/(ClusterName=="
            "'%3B%2F%3F%3A%40%20%26%3D%2B%24%2C')?group=None")
        clust.Cluster.search(self.adp, name=';/?:@ &=+$,')

    @mock.patch('pypowervm.adapter.Adapter.read')
    def test_search_by_feed(self, mock_read):
        """Test a search key that's not in search_keys."""
        def validate_read(root_type, xag):
            # This should be called by _search_by_feed, not by search.
            # Otherwise, we'll get an exception on the arg list.
            self.assertEqual(net.NetBridge.schema_type, root_type)
            self.assertIsNone(xag)
            return pvmhttp.load_pvm_resp(NET_BRIDGE_FILE).get_response()
        mock_read.side_effect = validate_read
        # vswitch_id is particularly cool because it's not just a top-level
        # child element - it dives into the SEAs, finds the primary trunk
        # adapter, and returns the vswitch ID from there.
        rets = net.NetBridge.search(self.adp, vswitch_id=0)
        self.assertEqual(1, len(rets))
        nb = rets[0]
        self.assertIsInstance(nb, net.NetBridge)
        self.assertEqual('d648eb60-4d39-34ad-ae2b-928d8c9577ad', nb.uuid)
        # Test one_result down the no-search-key path
        nb = net.NetBridge.search(self.adp, one_result=True, vswitch_id=0)
        self.assertEqual('d648eb60-4d39-34ad-ae2b-928d8c9577ad', nb.uuid)

        # Now do a search that returns more than one item.
        # Use a string for an int field to prove it works anyway.
        rets = net.NetBridge.search(self.adp, pvid='1')
        self.assertEqual(2, len(rets))
        self.assertIsInstance(rets[0], net.NetBridge)
        self.assertIsInstance(rets[1], net.NetBridge)
        self.assertEqual({'d648eb60-4d39-34ad-ae2b-928d8c9577ad',
                          '764f3423-04c5-3b96-95a3-4764065400bd'},
                         {nb.uuid for nb in rets})
        # Ensure one_result returns the first hit
        self.assertEqual(rets[0].uuid, net.NetBridge.search(
            self.adp, one_result=True, pvid=1).uuid)

    @mock.patch('pypowervm.adapter.Adapter.read')
    def test_search_with_xag(self, mock_read):
        """Test a search key that's in search_keys, but specifying xag."""
        def validate_read(root_type, xag):
            # This should be called by _search_by_feed, not by search.
            # Otherwise, we'll get an exception on the arg list.
            self.assertEqual(lpar.LPAR.schema_type, root_type)
            self.assertEqual(['Foo', 'Bar'], xag)
            return pvmhttp.load_pvm_resp(LPAR_FILE).get_response()
        mock_read.side_effect = validate_read
        rets = lpar.LPAR.search(self.adp, name='linux1', xag=['Foo', 'Bar'])
        self.assertEqual(1, len(rets))
        linux1 = rets[0]
        self.assertIsInstance(linux1, lpar.LPAR)
        self.assertEqual('9068B0FB-1CF0-4D23-8A23-31AC87D5F5D2', linux1.uuid)

    @mock.patch('pypowervm.adapter.Adapter.read')
    def test_child_no_search_key(self, mock_read):
        """CHILD search with or without a search key (uses GET-feed-loop)."""
        def validate_read(root_type, root_id, child_type, xag):
            # This should be called by _search_by_feed, not by search.
            # Otherwise, we'll get an exception on the arg list.
            self.assertEqual('SomeParent', root_type)
            self.assertEqual('some_uuid', root_id)
            self.assertEqual('Cluster', child_type)
            self.assertIsNone(xag)
            return pvmhttp.load_pvm_resp(NET_BRIDGE_FILE).get_response()
        mock_read.side_effect = validate_read
        clust.Cluster.search(self.adp, id=0, parent_type='SomeParent',
                             parent_uuid='some_uuid')
        clust.Cluster.search(self.adp, name='mycluster',
                             parent_type='SomeParent', parent_uuid='some_uuid')

    def test_child_bad_args(self):
        """Specifying parent_uuid without parent_type is an error."""
        self.assertRaises(ValueError, net.NetBridge.search, self.adp,
                          vswitch_id=0, parent_uuid='some_uuid')

    @mock.patch('pypowervm.adapter.Adapter.read')
    def test_search_all_parents(self, mock_read):
        """Anonymous ROOT for CHILD search."""
        # We're going to pretend that VSwitch is a CHILD of VirtualIOServer.
        parent_type = 'VirtualIOServer'
        child_schema_type = 'VirtualNetwork'

        # Anonymous CHILD search should call GET(ROOT feed), followed by one
        # GET(CHILD feed) for each ROOT parent.  Choosing a feed file with two
        # entries.  The following mock_reads are chained in order.
        def validate_feed_get(root):
            # Chain to the first entry GET
            mock_read.side_effect = validate_entry_get1
            self.assertEqual(parent_type, root)
            # VIOS_FILE has two <entry>s.
            return pvmhttp.load_pvm_resp(VIOS_FILE).get_response()

        def validate_entry_get1(root, root_id, child_type, **kwargs):
            # Chain to the second entry GET
            mock_read.side_effect = validate_entry_get2
            self.assertEqual(parent_type, root)
            self.assertEqual('1300C76F-9814-4A4D-B1F0-5B69352A7DEA', root_id)
            self.assertEqual(child_schema_type, child_type)
            entry_resp = pvmhttp.load_pvm_resp(VNETS_FILE).get_response()
            # Use the first half of the feed, which contains two tagged vlans
            entry_resp.feed.entries = entry_resp.feed.entries[:4]
            return entry_resp

        def validate_entry_get2(root, root_id, child_type, **kwargs):
            self.assertEqual(parent_type, root)
            self.assertEqual('7DBBE705-E4C4-4458-8223-3EBE07015CA9', root_id)
            self.assertEqual(child_schema_type, child_type)
            entry_resp = pvmhttp.load_pvm_resp(VNETS_FILE).get_response()
            # Use the second half of the feed, which contains two tagged vlans
            entry_resp.feed.entries = entry_resp.feed.entries[4:]
            return entry_resp

        # Set up the first mock_read in the chain
        mock_read.side_effect = validate_feed_get

        # Do the search (with class as parent_type)
        wraps = net.VNet.search(self.adp, parent_type=vios.VIOS, tagged=True)
        # Make sure we got the right networks
        self.assertEqual(4, len(wraps))
        for wrap, expected_vlanid in zip(wraps, (1234, 2, 1001, 1000)):
            self.assertIsInstance(wrap, net.VNet)
            self.assertTrue(wrap.tagged)
            self.assertEqual(expected_vlanid, wrap.vlan)

    @mock.patch('pypowervm.adapter.Adapter.read')
    def test_child_with_parent_spec(self, mock_read):
        """Test CHILD search using a parent instance."""
        def validate_read(root_type, root_id, child_type, xag):
            self.assertEqual('st', root_type)
            self.assertEqual('uuid', root_id)
            self.assertEqual('Cluster', child_type)
            self.assertIsNone(xag)
            return pvmhttp.load_pvm_resp(NET_BRIDGE_FILE).get_response()
        mock_read.side_effect = validate_read
        parent = mock.Mock(spec=clust.Cluster, schema_type='st', uuid='uuid')
        clust.Cluster.search(self.adp, id=0, parent=parent)


class TestRefresh(testtools.TestCase):
    """Tests for Adapter.refresh()."""

    clust_uuid = 'cluster_uuid'
    clust_href = 'https://server:12443/rest/api/uom/Cluster' + clust_uuid

    def setUp(self):
        super(TestRefresh, self).setUp()
        self.adp = apt.Adapter(mock.patch('requests.Session'))
        props = {'id': self.clust_uuid, 'links': {'SELF': [self.clust_href]}}
        self.old_etag = '123'
        self.clust_old = clust.Cluster.bld(
            self.adp, 'mycluster', stor.PV.bld(self.adp, 'hdisk1', 'udid1'),
            clust.Node.bld(self.adp, 'hostname1'))
        self.clust_old._etag = None
        self.clust_old.entry.properties = props
        self.new_etag = '456'
        self.clust_new = clust.Cluster.bld(
            self.adp, 'mycluster', stor.PV.bld(self.adp, 'hdisk2', 'udid2'),
            clust.Node.bld(self.adp, 'hostname2'))
        self.clust_new._etag = self.new_etag
        self.clust_new.entry.properties = props
        self.resp304 = apt.Response(
            'meth', 'path', 304, 'reason', {'etag': self.old_etag})
        self.resp200old = apt.Response(
            'meth', 'path', 200, 'reason', {'etag': self.old_etag})
        self.resp200old.entry = self.clust_old.entry
        self.resp200new = apt.Response(
            'meth', 'path', 200, 'reason', {'etag': self.new_etag})
        self.resp200new.entry = self.clust_new.entry

    def _mock_read_by_href(self, in_etag, out_resp):
        def read_by_href(href, etag, *args, **kwargs):
            self.assertEqual(href, self.clust_href)
            self.assertEqual(etag, in_etag)
            return out_resp
        return read_by_href

    @mock.patch('pypowervm.adapter.Adapter.read_by_href')
    def test_no_etag(self, mock_read):
        mock_read.side_effect = self._mock_read_by_href(
            None, self.resp200old)
        clust_old_save = copy.deepcopy(self.clust_old)
        clust_refreshed = self.clust_old.refresh()
        _assert_clusters_equal(self, clust_old_save, clust_refreshed)

    @mock.patch('pypowervm.adapter.Adapter.read_by_href')
    def test_etag_match(self, mock_read):
        mock_read.side_effect = self._mock_read_by_href(
            self.old_etag, self.resp304)
        self.clust_old._etag = self.old_etag
        clust_refreshed = self.clust_old.refresh()
        # On an etag match, refresh should return the same instance
        self.assertEqual(self.clust_old, clust_refreshed)

    @mock.patch('pypowervm.adapter.Adapter.read_by_href')
    def test_etag_no_match(self, mock_read):
        mock_read.side_effect = self._mock_read_by_href(
            self.old_etag, self.resp200new)
        self.clust_old._etag = self.old_etag
        clust_new_save = copy.deepcopy(self.clust_new)
        clust_refreshed = self.clust_old.refresh()
        _assert_clusters_equal(self, clust_new_save, clust_refreshed)

    @mock.patch('pypowervm.adapter.Adapter.read_by_href')
    def test_use_etag_false(self, mock_read):
        mock_read.side_effect = self._mock_read_by_href(
            None, self.resp200new)
        self.clust_old._etag = self.old_etag
        clust_new_save = copy.deepcopy(self.clust_new)
        clust_refreshed = self.clust_old.refresh(use_etag=False)
        _assert_clusters_equal(self, clust_new_save, clust_refreshed)


class TestUpdate(testtools.TestCase):
    clust_uuid = 'cluster_uuid'
    clust_path = '/rest/api/uom/Cluster' + clust_uuid
    clust_href = 'https://server:12443' + clust_path
    clust_etag = '123'

    def setUp(self):
        super(TestUpdate, self).setUp()
        self.adp = apt.Adapter(self.useFixture(fx.SessionFx()).sess)

        props = {'id': self.clust_uuid, 'links': {'SELF': [self.clust_href]}}
        self.cl = clust.Cluster.bld(
            self.adp, 'mycluster', stor.PV.bld(
                self.adp, 'hdisk1', udid='udid1'),
            clust.Node.bld(self.adp, 'hostname1'))
        self.cl._etag = self.clust_etag
        self.cl.entry.properties = props

    @mock.patch('pypowervm.adapter.Adapter.update_by_path')
    def test_update(self, mock_ubp):
        new_etag = '456'
        resp = apt.Response('meth', 'path', 200, 'reason', {'etag': new_etag})
        resp.entry = self.cl.entry
        mock_ubp.return_value = resp
        newcl = self.cl.update()
        mock_ubp.assert_called_with(
            self.cl, self.clust_etag, self.clust_path, timeout=3600)
        _assert_clusters_equal(self, self.cl, newcl)
        self.assertEqual(newcl.etag, new_etag)

    @mock.patch('pypowervm.adapter.Adapter.update_by_path')
    @mock.patch('warnings.warn')
    def test_update_xag(self, mock_warn, mock_ubp):
        new_etag = '456'
        resp = apt.Response('meth', 'path', 200, 'reason', {'etag': new_etag})
        resp.entry = self.cl.entry
        mock_ubp.return_value = resp
        newcl = self.cl.update(xag=['one', 'two', 'three'], timeout=123)
        mock_ubp.assert_called_with(
            self.cl, self.clust_etag, self.clust_path, timeout=123)
        _assert_clusters_equal(self, self.cl, newcl)
        self.assertEqual(newcl.etag, new_etag)
        mock_warn.assert_called_with(mock.ANY, DeprecationWarning)

    @mock.patch('pypowervm.adapter.Adapter.update_by_path')
    @mock.patch('warnings.warn')
    def test_update_with_get_xag(self, mock_warn, mock_ubp):
        # Update the entry with the new properties
        get_href = self.clust_href + "?group=one,three,two"
        props = {'id': self.clust_uuid, 'links': {'SELF': [get_href]}}
        self.cl.entry.properties = props

        new_etag = '456'
        resp = apt.Response('meth', 'path', 200, 'reason', {'etag': new_etag})
        resp.entry = self.cl.entry
        mock_ubp.return_value = resp
        newcl = self.cl.update(xag=['should', 'be', 'ignored'], timeout=-1)
        mock_ubp.assert_called_with(
            self.cl, self.clust_etag, self.clust_path + '?group=one,three,two',
            timeout=3600)
        _assert_clusters_equal(self, self.cl, newcl)
        self.assertEqual(newcl.etag, new_etag)
        mock_warn.assert_called_with(mock.ANY, DeprecationWarning)


class TestDelete(testtools.TestCase):
    def setUp(self):
        super(TestDelete, self).setUp()
        self.adpt = self.useFixture(fx.AdapterFx()).adpt

    def test_delete(self):
        vswitch = net.VSwitch.bld(self.adpt, 'a_switch')
        vswitch.entry = mock.MagicMock()
        vswitch.entry.href = 'test'
        vswitch.entry.etag = 5

        def validate_delete(uri, etag):
            self.assertEqual('test', uri)
            self.assertEqual(5, etag)
            return

        self.adpt.delete_by_href.side_effect = validate_delete
        vswitch.delete()


class TestCreate(testtools.TestCase):
    def setUp(self):
        super(TestCreate, self).setUp()
        self.adpt = self.useFixture(fx.AdapterFx()).adpt

    def test_create_root(self):
        vswitch = net.VSwitch.bld(self.adpt, 'a_switch')

        def validate_create(element, root_type, service, timeout=-1):
            self.assertIsInstance(element, net.VSwitch)
            self.assertEqual(net.VSwitch.schema_type, root_type)
            self.assertEqual('uom', service)
            self.assertEqual(123, timeout)
            return vswitch.entry
        self.adpt.create.side_effect = validate_create
        vswitch.create(timeout=123)

    def test_create_child(self):
        # We can safely pretend VSwitch is a child for purposes of this test.
        vswitch = net.VSwitch.bld(self.adpt, 'a_switch')

        def validate_create(element, root_type, root_id, child_type, service,
                            timeout=456):
            self.assertIsInstance(element, net.VSwitch)
            self.assertEqual(net.VSwitch.schema_type, child_type)
            self.assertEqual('NetworkBridge', root_type)
            self.assertEqual('SomeUUID', root_id)
            self.assertEqual('uom', service)
            self.assertEqual(-1, timeout)
            return vswitch.entry
        self.adpt.create.side_effect = validate_create
        # Make sure it works when parent_type is a class...
        vswitch.create(parent_type=net.NetBridge, parent_uuid='SomeUUID')
        # ...or a string
        vswitch.create(parent_type='NetworkBridge', parent_uuid='SomeUUID')
        # Or an instance
        parent = mock.Mock(spec=net.NetBridge, schema_type='NetworkBridge',
                           uuid='SomeUUID')
        vswitch.create(parent=parent)

    def test_create_other_service(self):
        """Ensure non-UOM service goes through."""
        vfile = vf.File.bld(self.adpt, "filename", "filetype", "vios_uuid")

        def validate_create(element, root_type, service, timeout=345):
            self.assertIsInstance(element, vf.File)
            self.assertEqual(vf.File.schema_type, root_type)
            self.assertEqual('web', service)
            self.assertEqual(-1, timeout)
            return vfile.entry
        self.adpt.create.side_effect = validate_create
        vfile.create(timeout=-1)

    def test_create_raises(self):
        """Verify invalid inputs raise exceptions."""
        vswitch = net.VSwitch.bld(self.adpt, 'a_switch')
        self.assertRaises(ValueError, vswitch.create, parent_type='Foo')
        self.assertRaises(ValueError, vswitch.create, parent_uuid='Foo')


class TestSetUUIDMixin(testtools.TestCase):
    """Generic tests for WrapperSetUUIDMixin."""
    def test_set_uuid(self):
        """Test mixins of Element (with/without Metadata) and Entry."""
        old_uuid = pvm_uuid.convert_uuid_to_pvm(str(uuid.uuid4()))
        new_uuid = pvm_uuid.convert_uuid_to_pvm(str(uuid.uuid4()))

        def set_wrap_uuid(wrap, value):
            wrap.uuid = value

        def assert_uuid(wrap, uuid, has_entry=True):
            if has_entry:
                self.assertEqual(uuid, wrap.uuid)
                # Same as above
                if uuid is None:
                    self.assertTrue('id' not in wrap.entry.properties or
                                    wrap.entry.properties['id'] is None)
                else:
                    self.assertEqual(uuid, wrap.entry.properties['id'])
            else:
                self.assertFalse(hasattr(wrap, 'entry'))
            self.assertEqual(uuid, wrap.uuid)

        @ewrap.EntryWrapper.pvm_type('SomeEntry')
        class SomeEntry(ewrap.EntryWrapper, ewrap.WrapperSetUUIDMixin):
            """EntryWrapper with set-uuid mixin."""
            pass

        # Set bad uuid value
        bad_uuid = 'F' + new_uuid[1:]
        self.assertRaises(ValueError, set_wrap_uuid,
                          SomeEntry(None), bad_uuid)

        # Entry has both Metadata and properties['id']
        some_ent = SomeEntry._bld(None)
        # Starts off empty
        assert_uuid(some_ent, None)
        # Can set from empty
        some_ent.set_uuid(old_uuid)
        assert_uuid(some_ent, old_uuid)
        # Can change from already-set
        some_ent.set_uuid(new_uuid)
        assert_uuid(some_ent, new_uuid)

        @ewrap.ElementWrapper.pvm_type('SomeObject', has_metadata=True)
        class SomeElementWithMetadata(ewrap.ElementWrapper,
                                      ewrap.WrapperSetUUIDMixin):
            """ElementWrapper with set-uuid mixin - WITH Metadata."""
            pass

        # Element has it in one place.  Also testing vivification of AtomID.
        sewm = SomeElementWithMetadata._bld(None)
        # Starts with no AtomID
        self.assertEqual(
            '<uom:SomeObject xmlns:uom="http://www.ibm.com/xmlns/systems/power'
            '/firmware/uom/mc/2012_10/" schemaVersion="V1_0"><uom:Metadata>'
            '<uom:Atom/></uom:Metadata></uom:SomeObject>'.encode('utf-8'),
            sewm.toxmlstring())
        assert_uuid(sewm, None, has_entry=False)
        # Can set
        sewm.set_uuid(old_uuid)
        assert_uuid(sewm, old_uuid, has_entry=False)
        # Can change
        sewm.set_uuid(new_uuid)
        assert_uuid(sewm, new_uuid, has_entry=False)

        @ewrap.ElementWrapper.pvm_type('SomeOtherObject')
        class SomeElementWithoutMetadata(ewrap.ElementWrapper,
                                         ewrap.WrapperSetUUIDMixin):
            """ElementWrapper with set-uuid mixin - WITHOUT Metadata."""
            pass

        sewom = SomeElementWithoutMetadata._bld(None)
        # No Metadata
        self.assertEqual(
            '<uom:SomeOtherObject xmlns:uom="http://www.ibm.com/xmlns/systems'
            '/power/firmware/uom/mc/2012_10/" schemaVersion="V1_0"/>'.
            encode('utf-8'),
            sewom.toxmlstring())
        assert_uuid(sewom, None, has_entry=False)
        # Exception attempting to set on an element with no metadata
        self.assertRaises(AttributeError, sewom.set_uuid, new_uuid)


class TestGetters(twrap.TestWrapper):
    file = LPAR_FILE
    wrapper_class_to_test = lpar.LPAR

    @mock.patch('pypowervm.wrappers.entry_wrapper.EntryWrapper.refresh')
    def test_entry_wrapper_getter(self, mock_refresh):
        self.adpt.read.return_value = self.dwrap.entry
        mock_refresh.return_value = self.dwrap
        # ROOT
        getter = ewrap.EntryWrapperGetter(self.adpt, lpar.LPAR, 'lpar_uuid')
        self.assertEqual('lpar_uuid', getter.uuid)
        lwrap = getter.get()
        self.assertIsInstance(lwrap, lpar.LPAR)
        self.assertEqual(self.dwrap.entry, lwrap.entry)
        self.adpt.read.assert_called_with(
            'LogicalPartition', 'lpar_uuid', child_id=None, child_type=None,
            xag=None)
        self.assertEqual(1, self.adpt.read.call_count)
        self.assertEqual(0, mock_refresh.call_count)
        # Second get doesn't re-read
        lwrap = getter.get()
        self.assertIsInstance(lwrap, lpar.LPAR)
        self.assertEqual(self.dwrap.entry, lwrap.entry)
        self.assertEqual(1, self.adpt.read.call_count)
        self.assertEqual(0, mock_refresh.call_count)
        # get with refresh doesn't read, but does refresh
        lwrap = getter.get(refresh=True)
        self.assertIsInstance(lwrap, lpar.LPAR)
        self.assertEqual(self.dwrap.entry, lwrap.entry)
        self.assertEqual(1, self.adpt.read.call_count)
        self.assertEqual(1, mock_refresh.call_count)

        # CHILD, use the EntryWrapper.getter classmethod, use xags
        getter = lpar.LPAR.getter(
            self.adpt, 'lpar_uuid', parent_class=stor.VDisk,
            parent_uuid='parent_uuid', xag=['one', 'two'])
        self.assertIsInstance(getter, ewrap.EntryWrapperGetter)
        self.assertEqual('lpar_uuid', getter.uuid)
        lwrap = getter.get()
        self.assertIsInstance(lwrap, lpar.LPAR)
        self.adpt.read.assert_called_with(
            'VirtualDisk', 'parent_uuid', child_type='LogicalPartition',
            child_id='lpar_uuid', xag=['one', 'two'])

        # With string parent_class
        getter = lpar.LPAR.getter(
            self.adpt, 'lpar_uuid', parent_class='VirtualDisk',
            parent_uuid='parent_uuid', xag=['one', 'two'])
        self.assertIsInstance(getter, ewrap.EntryWrapperGetter)
        self.assertEqual('lpar_uuid', getter.uuid)
        lwrap = getter.get()
        self.assertIsInstance(lwrap, lpar.LPAR)
        self.adpt.read.assert_called_with(
            'VirtualDisk', 'parent_uuid', child_type='LogicalPartition',
            child_id='lpar_uuid', xag=['one', 'two'])

        # With parent instance
        parent = mock.Mock(spec=stor.VDisk, schema_type='st', uuid='uuid')
        getter = lpar.LPAR.getter(self.adpt, 'lpar_uuid', parent=parent)
        self.assertIsInstance(getter, ewrap.EntryWrapperGetter)
        self.assertEqual('lpar_uuid', getter.uuid)
        lwrap = getter.get()
        self.assertIsInstance(lwrap, lpar.LPAR)
        self.adpt.read.assert_called_with(
            'st', 'uuid', child_type='LogicalPartition', child_id='lpar_uuid',
            xag=None)

        # parent type & uuid must both be specified
        self.assertRaises(ValueError, ewrap.EntryWrapperGetter, self.adpt,
                          lpar.LPAR, 'lpar_uuid', parent_class=stor.VDisk)
        self.assertRaises(ValueError, ewrap.EntryWrapperGetter, self.adpt,
                          lpar.LPAR, 'lpar_uuid', parent_uuid='parent_uuid')
        # entry_class must be a Wrapper subtype
        self.assertRaises(ValueError, ewrap.EntryWrapperGetter, self.adpt, 's',
                          'lpar_uuid')
        self.assertRaises(ValueError, ewrap.EntryWrapperGetter, self.adpt,
                          None, 'lpar_uuid')
        self.assertRaises(ValueError, ewrap.EntryWrapperGetter, self.adpt,
                          ewrap.EntryWrapperGetter, 'lpar_uuid')
        self.assertRaises(ValueError, ewrap.EntryWrapperGetter, self.adpt,
                          lpar.LPAR, 'lpar_uuid', parent_class=None,
                          parent_uuid='parent_uuid')
        self.assertRaises(ValueError, ewrap.EntryWrapperGetter, self.adpt,
                          lpar.LPAR, 'lpar_uuid',
                          parent_class=ewrap.EntryWrapperGetter,
                          parent_uuid='parent_uuid')

    @mock.patch('pypowervm.wrappers.entry_wrapper.EntryWrapper.refresh')
    def test_feed_getter(self, mock_refresh):
        self.adpt.read.return_value = self.resp
        feediter = iter(self.entries)
        mock_refresh.side_effect = lambda: next(feediter)
        # ROOT
        getter = ewrap.FeedGetter(self.adpt, lpar.LPAR)
        lfeed = getter.get()
        self.assertEqual(21, len(lfeed))
        self.assertEqual('089FFB20-5D19-4A8C-BB80-13650627D985', lfeed[0].uuid)
        self.adpt.read.assert_called_with(
            'LogicalPartition', None, child_id=None, child_type=None, xag=None)
        self.assertEqual(1, self.adpt.read.call_count)
        self.assertEqual(0, mock_refresh.call_count)
        # Second get doesn't re-read
        lfeed = getter.get()
        self.assertEqual(21, len(lfeed))
        self.assertEqual('089FFB20-5D19-4A8C-BB80-13650627D985', lfeed[0].uuid)
        self.assertEqual(1, self.adpt.read.call_count)
        self.assertEqual(0, mock_refresh.call_count)
        # get with refresh refreshes all 21 wrappers (but doesn't call read)
        lfeed = getter.get(refresh=True)
        self.assertEqual(21, len(lfeed))
        self.assertEqual('089FFB20-5D19-4A8C-BB80-13650627D985', lfeed[0].uuid)
        self.assertEqual(1, self.adpt.read.call_count)
        self.assertEqual(21, mock_refresh.call_count)
        # get with refetch calls read, not refresh
        lfeed = getter.get(refetch=True)
        self.assertEqual(21, len(lfeed))
        self.assertEqual('089FFB20-5D19-4A8C-BB80-13650627D985', lfeed[0].uuid)
        self.assertEqual(2, self.adpt.read.call_count)
        self.adpt.read.assert_called_with(
            'LogicalPartition', None, child_id=None, child_type=None, xag=None)
        self.assertEqual(21, mock_refresh.call_count)

        # CHILD, use the EntryWrapper.getter classmethod, use xags
        getter = lpar.LPAR.getter(self.adpt, parent_class=stor.VDisk,
                                  parent_uuid='p_uuid', xag=['one', 'two'])
        self.assertIsInstance(getter, ewrap.FeedGetter)
        lfeed = getter.get()
        self.assertEqual(21, len(lfeed))
        self.assertEqual('089FFB20-5D19-4A8C-BB80-13650627D985', lfeed[0].uuid)
        self.adpt.read.assert_called_with(
            'VirtualDisk', 'p_uuid', child_type='LogicalPartition',
            child_id=None, xag=['one', 'two'])

        # CHILD, parent_class as string schema type
        getter = lpar.LPAR.getter(self.adpt, parent_class='VirtualDisk',
                                  parent_uuid='p_uuid', xag=['one', 'two'])
        self.assertIsInstance(getter, ewrap.FeedGetter)
        lfeed = getter.get()
        self.assertEqual(21, len(lfeed))
        self.assertEqual('089FFB20-5D19-4A8C-BB80-13650627D985', lfeed[0].uuid)
        self.adpt.read.assert_called_with(
            'VirtualDisk', 'p_uuid', child_type='LogicalPartition',
            child_id=None, xag=['one', 'two'])

        # CHILD, parent instance
        parent = mock.Mock(spec=stor.VDisk, schema_type='st', uuid='uuid')
        getter = lpar.LPAR.getter(self.adpt, parent=parent)
        self.assertIsInstance(getter, ewrap.FeedGetter)
        lfeed = getter.get()
        self.assertEqual(21, len(lfeed))
        self.assertEqual('089FFB20-5D19-4A8C-BB80-13650627D985', lfeed[0].uuid)
        self.adpt.read.assert_called_with(
            'st', 'uuid', child_type='LogicalPartition', child_id=None,
            xag=None)

        # entry_class must be a Wrapper subtype
        self.assertRaises(ValueError, ewrap.FeedGetter, self.adpt, 's')
        self.assertRaises(ValueError, ewrap.FeedGetter, self.adpt, None)
        self.assertRaises(ValueError, ewrap.FeedGetter, self.adpt,
                          ewrap.EntryWrapperGetter)
        self.assertRaises(ValueError, ewrap.FeedGetter, self.adpt, lpar.LPAR,
                          parent_class=None, parent_uuid='parent_uuid')
        self.assertRaises(ValueError, ewrap.FeedGetter, self.adpt, lpar.LPAR,
                          parent_class=ewrap.EntryWrapperGetter,
                          parent_uuid='parent_uuid')

    @mock.patch('pypowervm.wrappers.entry_wrapper.EntryWrapper.refresh')
    def test_uuid_feed_getter(self, mock_refresh):
        """Verify UUIDFeedGetter."""
        # Mock return separate entries per read.  Need multiple copies for
        # multiple calls.
        read_iter = iter(wrp.entry for wrp in (
            self.entries[:3] + self.entries[:3] + self.entries[:3] +
            self.entries[:3]))
        self.adpt.read.side_effect = lambda *a, **k: next(read_iter)
        # Separate iterator for refreshes
        refresh_iter = iter(self.entries[:3])
        mock_refresh.side_effect = lambda: next(refresh_iter)
        # ROOT
        uuids = ['u1', 'u2', 'u3']
        getter = ewrap.UUIDFeedGetter(self.adpt, lpar.LPAR, uuids)
        # In order to be useful for a FeedTask, this has to evaluate as an
        # instance of FeedGetter
        self.assertIsInstance(getter, ewrap.FeedGetter)
        lfeed = getter.get()
        self.assertEqual(3, len(lfeed))
        self.assertEqual('089FFB20-5D19-4A8C-BB80-13650627D985', lfeed[0].uuid)
        # This does three separate reads
        self.adpt.read.assert_has_calls([mock.call(
            lpar.LPAR.schema_type, uuid, child_type=None, child_id=None,
            xag=None) for uuid in uuids])
        self.assertEqual(0, mock_refresh.call_count)
        # Second get doesn't re-read
        lfeed = getter.get()
        self.assertEqual(3, len(lfeed))
        self.assertEqual('089FFB20-5D19-4A8C-BB80-13650627D985', lfeed[0].uuid)
        self.assertEqual(3, self.adpt.read.call_count)
        self.assertEqual(0, mock_refresh.call_count)
        # get with refresh refreshes all thre wrappers (but doesn't call read)
        lfeed = getter.get(refresh=True)
        self.assertEqual(3, len(lfeed))
        self.assertEqual('089FFB20-5D19-4A8C-BB80-13650627D985', lfeed[0].uuid)
        self.assertEqual(3, self.adpt.read.call_count)
        self.assertEqual(3, mock_refresh.call_count)
        # get with refetch calls read, not refresh
        lfeed = getter.get(refetch=True)
        self.assertEqual(3, len(lfeed))
        self.assertEqual('089FFB20-5D19-4A8C-BB80-13650627D985', lfeed[0].uuid)
        self.assertEqual(6, self.adpt.read.call_count)
        self.assertEqual(3, mock_refresh.call_count)

        # CHILD
        getter = ewrap.UUIDFeedGetter(
            self.adpt, lpar.LPAR, uuids, parent_class=stor.VDisk,
            parent_uuid='p_uuid', xag=['one', 'two'])
        self.assertIsInstance(getter, ewrap.FeedGetter)
        lfeed = getter.get()
        self.assertEqual(3, len(lfeed))
        self.assertEqual('089FFB20-5D19-4A8C-BB80-13650627D985', lfeed[0].uuid)
        self.adpt.read.assert_has_calls([mock.call(
            stor.VDisk.schema_type, 'p_uuid', child_type=lpar.LPAR.schema_type,
            child_id=uuid, xag=['one', 'two']) for uuid in uuids])

        # With parent instance
        parent = mock.Mock(spec=stor.VDisk, schema_type='st', uuid='uuid')
        getter = ewrap.UUIDFeedGetter(
            self.adpt, lpar.LPAR, uuids, parent=parent)
        self.assertIsInstance(getter, ewrap.FeedGetter)
        lfeed = getter.get()
        self.assertEqual(3, len(lfeed))
        self.assertEqual('089FFB20-5D19-4A8C-BB80-13650627D985', lfeed[0].uuid)
        self.adpt.read.assert_has_calls(
            [mock.call('st', 'uuid', child_type=lpar.LPAR.schema_type,
                       child_id=uuid, xag=None) for uuid in uuids])


if __name__ == '__main__':
    unittest.main()
